"""
RepVGG, adapted from
https://github.com/DingXiaoH/RepVGG/blob/main/repvgg.py

Paper "RepVGG: Making VGG-style ConvNets Great Again", https://arxiv.org/abs/2101.03697
"""

# Reference license: MIT

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import SqueezeExcitation

from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class RepVggBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int,
        padding: int,
        groups: int,
        use_se: bool,
        reparameterized: bool,
    ) -> None:
        super().__init__()
        self.reparameterized = reparameterized
        self.in_channels = in_channels
        self.kernel_size = kernel_size
        self.groups = groups

        if use_se is True:
            self.se = SqueezeExcitation(out_channels, out_channels // 16)
        else:
            self.se = nn.Identity()

        self.activation = nn.ReLU()

        if reparameterized is True:
            self.reparam_conv = nn.Conv2d(
                in_channels,
                out_channels,
                kernel_size=(kernel_size, kernel_size),
                stride=(stride, stride),
                padding=(padding, padding),
                groups=groups,
                bias=True,
            )
        else:
            self.reparam_conv = None

            self.rbr_identity = None
            if out_channels == in_channels and stride == 1:
                self.rbr_identity = nn.BatchNorm2d(num_features=in_channels)

            self.conv_kxk = nn.Sequential()
            self.conv_kxk.add_module(
                "conv",
                nn.Conv2d(
                    in_channels=in_channels,
                    out_channels=out_channels,
                    kernel_size=(kernel_size, kernel_size),
                    stride=(stride, stride),
                    padding=(1, 1),
                    groups=self.groups,
                    bias=False,
                ),
            )
            self.conv_kxk.add_module("bn", nn.BatchNorm2d(num_features=out_channels))

            self.conv_1x1 = nn.Sequential()
            self.conv_1x1.add_module(
                "conv",
                nn.Conv2d(
                    in_channels=in_channels,
                    out_channels=out_channels,
                    kernel_size=(1, 1),
                    stride=(stride, stride),
                    padding=(0, 0),
                    groups=self.groups,
                    bias=False,
                ),
            )
            self.conv_1x1.add_module("bn", nn.BatchNorm2d(num_features=out_channels))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Reparameterized forward pass
        if self.reparam_conv is not None:
            return self.activation(self.se(self.reparam_conv(x)))

        # Multi-branched train-time forward pass
        if self.rbr_identity is None:
            x = self.conv_1x1(x) + self.conv_kxk(x)
        else:
            identity = self.rbr_identity(x)
            x = self.conv_1x1(x) + self.conv_kxk(x) + identity

        return self.activation(self.se(x))

    def reparameterize(self) -> None:
        if self.reparameterized is True:
            return

        (kernel, bias) = self._get_kernel_bias()
        self.reparam_conv = nn.Conv2d(
            in_channels=self.conv_kxk.conv.in_channels,
            out_channels=self.conv_kxk.conv.out_channels,
            kernel_size=self.conv_kxk.conv.kernel_size,
            stride=self.conv_kxk.conv.stride,
            padding=self.conv_kxk.conv.padding,
            dilation=self.conv_kxk.conv.dilation,
            groups=self.conv_kxk.conv.groups,
            bias=True,
        )
        self.reparam_conv.weight.data = kernel
        self.reparam_conv.bias.data = bias

        # Delete un-used branches
        for param in self.parameters():
            param.detach_()

        del self.conv_kxk
        del self.conv_1x1
        if hasattr(self, "rbr_identity") is True:
            del self.rbr_identity

        self.reparameterized = True

    def _get_kernel_bias(self) -> tuple[torch.Tensor, torch.Tensor]:
        # Get weights and bias of scale branch
        kernel_1x1 = 0
        bias_1x1 = 0
        if self.conv_1x1 is not None:
            kernel_1x1, bias_1x1 = self._fuse_bn_tensor(self.conv_1x1)
            pad = self.kernel_size // 2
            kernel_1x1 = F.pad(kernel_1x1, [pad, pad, pad, pad])

        # Get weights and bias of skip branch
        kernel_identity = 0
        bias_identity = 0
        if self.rbr_identity is not None:
            (kernel_identity, bias_identity) = self._fuse_bn_tensor(self.rbr_identity)

        # Get weights and bias of conv branches
        (kernel_conv, bias_conv) = self._fuse_bn_tensor(self.conv_kxk)

        kernel_final = kernel_conv + kernel_1x1 + kernel_identity
        bias_final = bias_conv + bias_1x1 + bias_identity

        return (kernel_final, bias_final)

    def _fuse_bn_tensor(self, branch: nn.Module) -> tuple[torch.Tensor, torch.Tensor]:
        if isinstance(branch, nn.Sequential):
            kernel = branch.conv.weight
            running_mean = branch.bn.running_mean
            running_var = branch.bn.running_var
            gamma = branch.bn.weight
            beta = branch.bn.bias
            eps = branch.bn.eps

        else:
            assert isinstance(branch, nn.BatchNorm2d)
            input_dim = self.in_channels // self.groups
            kernel_value = torch.zeros(
                (self.in_channels, input_dim, self.kernel_size, self.kernel_size),
                dtype=branch.weight.dtype,
                device=branch.weight.device,
            )
            for i in range(self.in_channels):
                kernel_value[i, i % input_dim, self.kernel_size // 2, self.kernel_size // 2] = 1

            kernel = kernel_value
            running_mean = branch.running_mean
            running_var = branch.running_var
            gamma = branch.weight
            beta = branch.bias
            eps = branch.eps

        std = (running_var + eps).sqrt()
        t = (gamma / std).reshape(-1, 1, 1, 1)

        return (kernel * t, beta - running_mean * gamma / std)


class RepVggStage(nn.Sequential):
    def __init__(
        self,
        in_planes: int,
        planes: int,
        num_blocks: int,
        groups: int,
        prev_blocks: int,
        use_se: bool,
        reparameterized: bool,
    ) -> None:
        super().__init__()
        strides = [2] + [1] * (num_blocks - 1)
        for idx, stride in enumerate(strides):
            if (prev_blocks + idx) < 27 and (prev_blocks + idx) % 2 == 0:
                current_groups = groups

            else:
                current_groups = 1

            self.append(
                RepVggBlock(
                    in_channels=in_planes,
                    out_channels=planes,
                    kernel_size=3,
                    stride=stride,
                    padding=1,
                    groups=current_groups,
                    use_se=use_se,
                    reparameterized=reparameterized,
                )
            )
            in_planes = planes


class RepVgg(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        self.reparameterized = False
        widths = [64, 128, 256, 512]
        width_multipliers: list[float] = self.config["width_multipliers"]
        num_blocks_per_stage: list[int] = self.config["num_blocks_per_stage"]
        groups: int = self.config["groups"]
        use_se: bool = self.config["use_se"]

        in_planes = min(64, int(widths[0] * width_multipliers[0]))

        self.stem = RepVggBlock(
            in_channels=self.input_channels,
            out_channels=in_planes,
            kernel_size=3,
            stride=2,
            padding=1,
            groups=1,
            use_se=False,
            reparameterized=self.reparameterized,
        )

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        prev_blocks = 1  # Due to stem
        for idx in range(len(num_blocks_per_stage)):  # pylint: disable=consider-using-enumerate
            out_planes = int(widths[idx] * width_multipliers[idx])
            stages[f"stage{idx+1}"] = RepVggStage(
                in_planes,
                out_planes,
                num_blocks=num_blocks_per_stage[idx],
                groups=groups,
                prev_blocks=prev_blocks,
                use_se=use_se,
                reparameterized=self.reparameterized,
            )
            return_channels.append(out_planes)

            in_planes = out_planes
            prev_blocks += num_blocks_per_stage[idx]

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = int(widths[-1] * width_multipliers[3])
        self.classifier = self.create_classifier()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def reparameterize_model(self) -> None:
        for module in self.modules():
            if hasattr(module, "reparameterize") is True:
                module.reparameterize()

        self.reparameterized = True


registry.register_model_config(
    "repvgg_a0",
    RepVgg,
    config={
        "width_multipliers": [0.75, 0.75, 0.75, 2.5],
        "num_blocks_per_stage": [2, 4, 14, 1],
        "groups": 1,
        "use_se": False,
    },
)
registry.register_model_config(
    "repvgg_a1",
    RepVgg,
    config={"width_multipliers": [1, 1, 1, 2.5], "num_blocks_per_stage": [2, 4, 14, 1], "groups": 1, "use_se": False},
)
registry.register_model_config(
    "repvgg_a2",
    RepVgg,
    config={
        "width_multipliers": [1.5, 1.5, 1.5, 2.75],
        "num_blocks_per_stage": [2, 4, 14, 1],
        "groups": 1,
        "use_se": False,
    },
)
registry.register_model_config(
    "repvgg_b0",
    RepVgg,
    config={"width_multipliers": [1, 1, 1, 2.5], "num_blocks_per_stage": [4, 6, 16, 1], "groups": 1, "use_se": False},
)
registry.register_model_config(
    "repvgg_b1",
    RepVgg,
    config={"width_multipliers": [2, 2, 2, 4], "num_blocks_per_stage": [4, 6, 16, 1], "groups": 1, "use_se": False},
)
registry.register_model_config(
    "repvgg_b1g4",
    RepVgg,
    config={"width_multipliers": [2, 2, 2, 4], "num_blocks_per_stage": [4, 6, 16, 1], "groups": 4, "use_se": False},
)
registry.register_model_config(
    "repvgg_b2",
    RepVgg,
    config={
        "width_multipliers": [2.5, 2.5, 2.5, 5],
        "num_blocks_per_stage": [4, 6, 16, 1],
        "groups": 1,
        "use_se": False,
    },
)
registry.register_model_config(
    "repvgg_b2g4",
    RepVgg,
    config={
        "width_multipliers": [2.5, 2.5, 2.5, 5],
        "num_blocks_per_stage": [4, 6, 16, 1],
        "groups": 4,
        "use_se": False,
    },
)
registry.register_model_config(
    "repvgg_b3",
    RepVgg,
    config={"width_multipliers": [3, 3, 3, 5], "num_blocks_per_stage": [4, 6, 16, 1], "groups": 1, "use_se": False},
)
registry.register_model_config(
    "repvgg_b3g4",
    RepVgg,
    config={"width_multipliers": [3, 3, 3, 5], "num_blocks_per_stage": [4, 6, 16, 1], "groups": 4, "use_se": False},
)
registry.register_model_config(
    "repvgg_d2se",
    RepVgg,
    config={
        "width_multipliers": [2.5, 2.5, 2.5, 5],
        "num_blocks_per_stage": [8, 14, 24, 1],
        "groups": 1,
        "use_se": True,
    },
)



from typing import NamedTuple
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn

FGSMResponse = NamedTuple(
    "FGSMResponse", [("out", torch.Tensor), ("perturbation", torch.Tensor), ("adv_out", torch.Tensor)]
)


class FGSM:
    def __init__(self, net: nn.Module, eps: float) -> None:
        self.net = net.eval()
        self.eps = eps

    def __call__(self, input_tensor: torch.Tensor, target: Optional[torch.Tensor]) -> FGSMResponse:
        input_tensor.requires_grad = True
        out = self.net(input_tensor)
        if target is None:
            target = torch.argmax(out, dim=1)

        loss = F.nll_loss(out, target)
        self.net.zero_grad()
        loss.backward()

        input_grad = input_tensor.grad.data
        sign_data_grad = input_grad.sign()
        perturbed_image = input_tensor + self.eps * sign_data_grad

        adv_out = self.net(perturbed_image)

        return FGSMResponse(F.softmax(out, dim=1), self.eps * sign_data_grad, F.softmax(adv_out, dim=1))



import argparse
import logging
from typing import Any

import torch

from birder.common import cli
from birder.common import fs_ops

logger = logging.getLogger(__name__)


class Ensemble(torch.nn.Module):
    def __init__(self, module_list: torch.nn.ModuleList) -> None:
        super().__init__()
        self.module_list = module_list

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        outs = []
        for module in self.module_list:
            outs.append(module(x))

        x = torch.stack(outs)
        x = torch.mean(x, dim=0)

        return x


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "ensemble-model",
        allow_abbrev=False,
        help="create an ensemble model from multiple models",
        description="create an ensemble model from multiple models",
        epilog=(
            "Usage examples:\n"
            "python -m birder.tools ensemble-model --networks convnext_v2_4_0 focalnet_3_0 "
            "swin_transformer_v2_1_0 --pts\n"
            "python -m birder.tools ensemble-model --networks mobilevit_v2_1_5_intermediate_80 "
            "edgevit_2_intermediate_100 --pt2"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument("--networks", type=str, required=True, nargs="+", help="networks to ensemble")
    subparser.add_argument("--force", action="store_true", help="override existing model")

    format_group = subparser.add_mutually_exclusive_group(required=True)
    format_group.add_argument("--pts", default=False, action="store_true", help="ensemble TorchScript models")
    format_group.add_argument("--pt2", default=False, action="store_true", help="ensemble pt2 models")

    subparser.set_defaults(func=main)


def main(args: argparse.Namespace) -> None:
    device = torch.device("cpu")
    nets = torch.nn.ModuleList()
    class_to_idx_list = []
    signature_list = []
    rgb_stats_list = []
    for network in args.networks:
        (net, model_info) = fs_ops.load_model(device, network, inference=True, pts=args.pts, pt2=args.pt2)
        nets.append(net)
        class_to_idx_list.append(model_info.class_to_idx)
        signature_list.append(model_info.signature)
        rgb_stats_list.append(model_info.rgb_stats)

    # Ensure all have the same class to index definitions
    if [class_to_idx_list[0]] * len(class_to_idx_list) != class_to_idx_list:
        raise ValueError("All networks must have the same class to index definition")

    if [signature_list[0]] * len(signature_list) != signature_list:
        logger.warning(f"Networks signatures differ, using signature={signature_list[0]}")

    if [rgb_stats_list[0]] * len(rgb_stats_list) != rgb_stats_list:
        logger.warning(f"Networks rgb values differ, using rgb values of {rgb_stats_list[0]}")

    signature = signature_list[0]
    class_to_idx = class_to_idx_list[0]
    rgb_stats = rgb_stats_list[0]
    ensemble = Ensemble(nets)

    network_name = "ensemble"
    model_path = fs_ops.model_path(network_name, pts=args.pts, pt2=args.pt2)
    if model_path.exists() is True and args.force is False:
        logger.warning("Ensemble already exists... aborting")
        raise SystemExit(1)

    logger.info(f"Saving model {model_path}...")
    if args.pt2 is True:
        signature["inputs"][0]["data_shape"][0] = 2  # Set batch size
        sample_shape = signature["inputs"][0]["data_shape"]
        batch_dim = torch.export.Dim("batch", min=1)
        exported_net = torch.export.export(
            ensemble, (torch.randn(*sample_shape, device=device),), dynamic_shapes={"x": {0: batch_dim}}
        )

        # Save model
        fs_ops.save_pt2(exported_net, model_path, nets[0].task, class_to_idx, signature, rgb_stats)

    elif args.pts is True:
        scripted_ensemble = torch.jit.script(ensemble)

        # Save model
        fs_ops.save_pts(scripted_ensemble, model_path, nets[0].task, class_to_idx, signature, rgb_stats)



import argparse
import json
import logging
from collections import Counter
from pathlib import Path
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import torch
from torch.utils.data import DataLoader
from torchvision.datasets import ImageFolder
from torchvision.io import decode_image
from torchvision.transforms import v2
from tqdm import tqdm

from birder.common import cli
from birder.conf import settings

logger = logging.getLogger(__name__)


def detection_object_count(directory: Path) -> tuple[Counter[str], int]:
    """
    Assumes directory is in LabelMe format
    """

    file_count = 0
    detection_count: Counter[str] = Counter()
    for d in directory.iterdir():
        if d.is_dir() is False:
            continue

        for file_path in d.glob("*.json"):
            with open(file_path, "r", encoding="utf=8") as handle:
                raw_label = json.load(handle)

            if raw_label["flags"].get("unknown", False) is True:
                # Don't count images with unknown species
                continue

            for shape in raw_label["shapes"]:
                detection_count.update([shape["label"]])

            file_count += 1

    return (detection_count, file_count)


def directory_label_count(directory: str | Path) -> Counter[str]:
    dataset = ImageFolder(directory)
    labels = [dataset.classes[sample[1]] for sample in dataset.samples]
    return Counter(labels)


def class_graph(args: argparse.Namespace) -> None:
    if args.detection_class_graph is True:
        (label_count, _) = detection_object_count(Path(args.data_path))
    else:
        label_count = directory_label_count(args.data_path)

    total_count = sum(label_count.values())
    (sorted_classes, sorted_count) = list(zip(*label_count.most_common()))

    fig = plt.figure()
    ax = fig.add_subplot()
    ax.barh(sorted_classes, sorted_count, color=plt.get_cmap("RdYlGn")(sorted_count))
    for i, count in enumerate(sorted_count):
        ax.text(
            count + 3,
            i,
            str(count),
            color="dimgrey",
            va="center",
            fontsize="small",
            fontweight="light",
            clip_on=True,
        )

    ax.set_title(
        f"{total_count:n} samples, {len(sorted_classes)} classes "
        f"({total_count / len(sorted_classes):.0f} samples per class on average)"
    )

    plt.show()


def set_size(_args: argparse.Namespace) -> None:
    training_dataset = ImageFolder(settings.TRAINING_DATA_PATH)
    validation_dataset = ImageFolder(settings.VALIDATION_DATA_PATH)
    testing_dataset = ImageFolder(settings.TESTING_DATA_PATH)
    logger.info(
        f"Training:   {len(training_dataset):,} samples with {len(training_dataset.classes)} classes "
        f"(average of {round(len(training_dataset) / len(training_dataset.classes))} per class)"
    )
    logger.info(
        f"Validation: {len(validation_dataset):,} samples with {len(validation_dataset.classes)} classes "
        f"(average of {round(len(validation_dataset) / len(validation_dataset.classes))} per class)"
    )
    logger.info(
        f"Testing:    {len(testing_dataset):,} samples with {len(testing_dataset.classes)} classes "
        f"(average of {round(len(testing_dataset) / len(testing_dataset.classes))} per class)"
    )
    logger.info(
        f"Total of {len(training_dataset) + len(validation_dataset) + len(testing_dataset):,} " "classification samples"
    )

    logger.info("---")
    (training_detection_count, training_file_count) = detection_object_count(
        settings.TRAINING_DETECTION_ANNOTATIONS_PATH
    )
    if training_file_count > 0:
        logger.info(
            f"Detection training:   {training_file_count:,} samples containing "
            f"{sum(training_detection_count.values()):,} "
            f"objects with {len(training_detection_count.keys())} classes (average of "
            f"{sum(training_detection_count.values()) / training_file_count:.1f} objects per sample)"
        )

    (validation_detection_count, validation_file_count) = detection_object_count(
        settings.VALIDATION_DETECTION_ANNOTATIONS_PATH
    )
    if validation_file_count > 0:
        logger.info(
            f"Detection validation: {validation_file_count:,} samples containing "
            f"{sum(validation_detection_count.values()):,} "
            f"objects with {len(validation_detection_count.keys())} classes (average of "
            f"{sum(validation_detection_count.values()) / validation_file_count:.1f} objects per sample)"
        )

    logger.info(f"Total of {training_file_count + validation_file_count:,} detection samples")

    logger.info("---")
    weak_dataset = ImageFolder(settings.WEAKLY_LABELED_DATA_PATH)
    logger.info(
        f"Weakly labeled: {len(weak_dataset):,} samples with {len(weak_dataset.classes)} classes "
        f"(average of {round(len(weak_dataset) / len(weak_dataset.classes))} per class)"
    )
    weak_val_dataset = ImageFolder(settings.WEAKLY_VAL_LABELED_DATA_PATH)
    logger.info(
        f"Weakly labeled validation: {len(weak_val_dataset):,} samples with {len(weak_val_dataset.classes)} classes "
        f"(average of {round(len(weak_val_dataset) / len(weak_val_dataset.classes))} per class)"
    )

    logger.info("---")
    num_classes = len(set(training_dataset.class_to_idx.keys()).union(set(weak_dataset.class_to_idx.keys())))
    logger.info(f"Total of {num_classes} unique classes")


def mean_and_std(args: argparse.Namespace) -> None:
    dataset = ImageFolder(
        args.data_path,
        transform=v2.Compose(
            [
                v2.Resize((256, 256), interpolation=v2.InterpolationMode.BILINEAR),
                v2.ToDtype(torch.float32, scale=True),
            ]
        ),
        loader=decode_image,
    )
    batch_size = 64
    data_loader = DataLoader(
        dataset,
        batch_size=batch_size,
        num_workers=8,
    )
    num_samples = len(dataset)
    mean = np.zeros(3, dtype=np.float64)
    mean_squares = np.zeros(3, dtype=np.float64)
    with tqdm(total=num_samples, initial=0, unit="images", unit_scale=True, leave=True) as progress:
        for images, _ in data_loader:
            mean += images.mean(dim=(2, 3)).sum(dim=0).numpy().astype(np.float64) / num_samples
            mean_squares += images.square().mean(dim=(2, 3)).sum(dim=0).numpy().astype(np.float64) / num_samples

            progress.update(n=batch_size)

    std = np.sqrt(mean_squares - np.square(mean))
    rgb_stats = {"mean": mean.round(decimals=4).tolist(), "std": std.round(decimals=4).tolist()}
    logger.info(f"mean: {rgb_stats['mean']}")
    logger.info(f"std: {rgb_stats['std']}")


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "stats",
        allow_abbrev=False,
        help="show image directory statistics",
        description="show image directory statistics",
        epilog=(
            "Usage examples:\n"
            "python -m birder.tools stats --class-graph\n"
            "python -m birder.tools stats --class-graph --data-path data/validation\n"
            "python -m birder.tools stats --detection-class-graph --data-path "
            "data/detection_data/training_annotations\n"
            "python -m birder.tools stats --set-size\n"
            "python -m birder.tools stats --rgb\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument("--class-graph", default=False, action="store_true", help="show class sample distribution")
    subparser.add_argument(
        "--detection-class-graph",
        default=False,
        action="store_true",
        help="show class sample distribution for detection directory",
    )
    subparser.add_argument("--set-size", default=False, action="store_true", help="show sample count per set")
    subparser.add_argument(
        "--rgb",
        default=False,
        action="store_true",
        help="calculate rgb mean and std",
    )
    subparser.add_argument("--data-path", type=str, default=settings.TRAINING_DATA_PATH, help="image directory")
    subparser.set_defaults(func=main)


def main(args: argparse.Namespace) -> None:
    if args.class_graph is True or args.detection_class_graph:
        class_graph(args)

    if args.set_size is True:
        set_size(args)

    if args.rgb is True:
        mean_and_std(args)



import itertools
import logging
from functools import partial
from typing import Optional

import matplotlib.axes
import matplotlib.figure
import matplotlib.gridspec
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import polars as pl
import torch
from sklearn.metrics import auc
from sklearn.metrics import average_precision_score
from sklearn.metrics import precision_recall_curve
from sklearn.metrics import roc_curve
from sklearn.preprocessing import label_binarize
from torchvision.datasets.folder import pil_loader
from torchvision.io import decode_image
from torchvision.utils import draw_bounding_boxes

from birder.conf import settings
from birder.results.classification import Results

logger = logging.getLogger(__name__)


def show_detections(
    image_path: str,
    detection: dict[str, torch.Tensor],
    class_to_idx: dict[str, int],
    score_threshold: float = 0.5,
    color_list: Optional[list[tuple[int, ...]]] = None,
    show: bool = True,
) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    img = decode_image(image_path)
    idx_to_class = dict(zip(class_to_idx.values(), class_to_idx.keys()))

    scores = detection["scores"]
    idxs = torch.where(scores > score_threshold)
    scores = scores[idxs]
    boxes = detection["boxes"][idxs]
    labels = detection["labels"][idxs]
    label_names = [f"{idx_to_class[i.item()]}: {s:.4f}" for i, s in zip(labels, scores)]
    if color_list is not None:
        colors = [color_list[label] for label in labels]
    else:
        colors = None

    if boxes.size(0) == 0:
        result_with_boxes = img
    else:
        result_with_boxes = draw_bounding_boxes(
            image=img,
            boxes=boxes,
            labels=label_names,
            colors=colors,
            width=3,
            font="DejaVuSans",
            font_size=14,
        )

    fig = plt.figure(num=image_path, figsize=(12, 9))
    ax = fig.add_subplot(1, 1, 1)
    ax.imshow(np.transpose(result_with_boxes, [1, 2, 0]))
    ax.axis("off")

    if show is True:
        plt.tight_layout()
        plt.show()

    return (fig, ax)


def show_top_k(
    image_path: str,
    out: npt.NDArray[np.float32],
    class_to_idx: dict[str, int],
    label: Optional[int | str] = None,
    show: bool = True,
) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    img = pil_loader(image_path)

    idx_to_class = dict(zip(class_to_idx.values(), class_to_idx.keys()))
    probabilities: list[float] = []
    predicted_class_names: list[str] = []

    logger.debug(f"'{image_path}'")
    for idx in np.argsort(out)[::-1][: settings.TOP_K]:
        probabilities.append(out[idx])
        predicted_class_names.append(idx_to_class[idx])
        logger.debug(f"{idx_to_class[idx]:<25}: {out[idx]:.4f}")

    logger.debug("---")

    fig = plt.figure(num=image_path)
    gs = matplotlib.gridspec.GridSpec(2, 1, height_ratios=[2, 1])

    ax = fig.add_subplot(gs[0, 0])
    ax.imshow(img)
    ax.axis("off")

    ax = fig.add_subplot(gs[1, 0])
    y_pos = np.arange(settings.TOP_K)
    bars = ax.barh(y_pos, probabilities, alpha=0.4, align="center")
    ax.set_yticks(y_pos)
    ax.set_yticklabels(predicted_class_names)
    ax.invert_yaxis()
    ax.set_xlim(0, 1)

    if label is not None and label != settings.NO_LABEL:
        # 'label' could be any int-like object like NumPy, Torch, etc.
        if not isinstance(label, str):
            label = idx_to_class[label]

        for idx, class_name in enumerate(predicted_class_names):
            if label == class_name:
                bars[idx].set_color("green")

    if show is True:
        plt.tight_layout()
        plt.show()

    return (fig, ax)


class ConfusionMatrix:
    def __init__(self, results: Results):
        self.results = results

    def save(self, path: str) -> None:
        """
        Save confusion matrix as a CSV file
        """

        cnf_df = pl.DataFrame(
            {
                "": self.results.label_names,
                **{name: self.results.confusion_matrix[:, i] for i, name in enumerate(self.results.label_names)},
            }
        )
        logger.info(f"Saving confusion matrix at {path}")
        cnf_df.write_csv(path)

    def show(self) -> None:
        """
        Show confusion matrix as matplotlib figure
        """

        # Define figure and axes
        fig = plt.figure(figsize=(12, 10))
        ax = fig.add_subplot(1, 1, 1)

        # Plot confusion matrix
        cnf_matrix = self.results.confusion_matrix
        cnf_ax = ax.imshow(cnf_matrix, interpolation="nearest", cmap=plt.get_cmap("Blues"))
        ax.set_title(f"Confusion matrix, accuracy {self.results.accuracy:.4f} on {len(self.results)} samples")
        plt.colorbar(cnf_ax)
        tick_marks = np.arange(len(self.results.unique_labels))
        class_names = [self.results.label_names[label_idx] for label_idx in self.results.unique_labels]

        # Set axis
        ax.set_xticks(tick_marks)
        ax.set_xticklabels(class_names, minor=False)
        plt.setp(ax.xaxis.get_majorticklabels(), rotation=90)
        ax.set_yticks(tick_marks)
        ax.set_yticklabels(class_names, minor=False)

        # Add text (matrix values)
        threshold = cnf_matrix.max() / 2.0
        for i, j in itertools.product(range(cnf_matrix.shape[0]), range(cnf_matrix.shape[1])):
            if i == j and np.sum(cnf_matrix, axis=1)[i] == cnf_matrix[i, j]:
                text_color = "greenyellow"

            elif i == j and (np.sum(cnf_matrix, axis=1)[i] * 0.75) > cnf_matrix[i, j]:
                text_color = "red"

            elif i == j and (np.sum(cnf_matrix, axis=1)[i] * 0.9) > cnf_matrix[i, j]:
                text_color = "orange"

            elif cnf_matrix[i, j] > threshold:
                text_color = "white"

            else:
                text_color = "black"

            ax.text(
                j,
                i,
                cnf_matrix[i, j],
                verticalalignment="center",
                horizontalalignment="center",
                color=text_color,
                fontsize="small",
                clip_on=True,
            )

        offset = 0.5
        (height, width) = cnf_matrix.shape
        ax.hlines(
            y=np.arange(height + 1) - offset,
            xmin=-offset,
            xmax=width - offset,
            linestyles="dashed",
            colors="grey",
            linewidth=0.5,
        )
        ax.vlines(
            x=np.arange(width + 1) - offset,
            ymin=-offset,
            ymax=height - offset,
            linestyles="dashed",
            colors="grey",
            linewidth=0.5,
        )

        plt.tight_layout()
        plt.show()


class ROC:
    def __init__(self, result: Optional[Results] = None) -> None:
        self.results_dict: dict[str, Results] = {}
        if result is not None:
            self.results_dict["result"] = result

    def add_result(self, name: str, result: Results) -> None:
        self.results_dict[name] = result

    def show(self, roc_classes: Optional[list[str]] = None) -> None:
        """
        Show roc curve as matplotlib figure
        """

        if roc_classes is None:
            roc_classes = []

        # Define figure and axes
        fig = plt.figure(figsize=(12, 10))
        ax = fig.add_subplot(1, 1, 1)

        for name, results in self.results_dict.items():
            fpr = {}
            tpr = {}
            roc_auc = {}
            for i in results.unique_labels:
                binary_labels = results.labels == i
                (fpr[i], tpr[i], _) = roc_curve(binary_labels, results.output[:, i])
                if np.sum(binary_labels) == 0:
                    tpr[i] = np.zeros_like(fpr[i])

                roc_auc[i] = auc(fpr[i], tpr[i])

            all_fpr = np.unique(np.concatenate([fpr[i] for i in results.unique_labels]))
            mean_tpr = np.zeros_like(all_fpr)
            for i in results.unique_labels:
                mean_tpr += np.interp(all_fpr, fpr[i], tpr[i])

            mean_tpr /= len(results.unique_labels)
            macro_auc = auc(all_fpr, mean_tpr)

            for cls in roc_classes:
                i = results.label_names.index(cls)
                ax.plot(fpr[i], tpr[i], label=f"{name} ROC curve for {cls} ({roc_auc[i]:.4f})")

            ax.plot(
                all_fpr,
                mean_tpr,
                linestyle=":",
                label=f"{name} Macro-average ROC curve ({macro_auc:.4f})",
            )

        ax.plot([0, 1], [0, 1], color="navy", lw=2, linestyle="--")
        ax.set_xlim((0.0, 1.0))
        ax.set_ylim((0.0, 1.05))
        ax.set_xlabel("False Positive Rate")
        ax.set_ylabel("True Positive Rate")
        ax.set_title("Receiver operating characteristic")
        plt.legend(loc="lower right")
        plt.tight_layout()
        plt.show()


class PrecisionRecall:
    def __init__(self, result: Optional[Results] = None) -> None:
        self.results_dict: dict[str, Results] = {}
        if result is not None:
            self.results_dict["result"] = result

    def add_result(self, name: str, result: Results) -> None:
        self.results_dict[name] = result

    def show(self, pr_classes: Optional[list[str]] = None) -> None:
        """
        Show precision recall curve as matplotlib figure
        """

        if pr_classes is None:
            pr_classes = []

        # Define figure and axes
        fig = plt.figure(figsize=(12, 10))
        ax = fig.add_subplot(1, 1, 1)

        legend_lines = []
        legend_labels: list[str] = []
        for name, results in self.results_dict.items():
            labels = label_binarize(results.labels, classes=range(len(results.label_names)))

            # A "micro-average" quantifying score on all classes jointly
            (precision, recall, _) = precision_recall_curve(labels.ravel(), results.output.ravel())
            average_precision = average_precision_score(labels.ravel(), results.output.ravel(), average="micro")

            line = ax.step(recall, precision, linestyle=":", where="post")
            legend_lines.append(line[0])
            legend_labels.append(f"{name} micro-average precision-recall ({average_precision:.4f})")

            # Per selected class
            for cls in pr_classes:
                i = results.label_names.index(cls)
                (precision, recall, _) = precision_recall_curve(labels[:, i], results.output[:, i])
                average_precision = average_precision_score(labels[:, i], results.output[:, i])
                line = ax.plot(recall, precision, lw=2)
                legend_lines.append(line[0])
                legend_labels.append(f"{name} precision-recall for class {cls} ({average_precision:.4f})")

        # iso-f1 curves
        f_scores = np.array([0.2, 0.4, 0.6, 0.8, 0.9])
        for f_score in f_scores:
            x = np.linspace(0.01, 1)
            y = f_score * x / (2 * x - f_score)
            line = ax.plot(x[y >= 0], y[y >= 0], color="gray", alpha=0.2)
            ax.annotate(f"f1={f_score:0.1f}", xy=(0.9, y[45] + 0.02))

        legend_lines.append(line[0])
        legend_labels.append("iso-f1 curves")

        ax.set_xlabel("Recall")
        ax.set_ylabel("Precision")
        ax.set_ylim((0.0, 1.05))
        ax.set_xlim((0.0, 1.0))
        ax.set_title("Precision Recall Curves")
        plt.legend(legend_lines, legend_labels, loc="lower left")
        plt.tight_layout()
        plt.show()


class ProbabilityHistogram:
    def __init__(self, results: Results) -> None:
        self.results = results

    def show(self, cls_a: str, cls_b: str) -> None:
        results_df = self.results.to_dataframe()
        hist = partial(np.histogram, bins=20, range=(0, 1), density=True)

        cls_a_df = results_df.filter(pl.col("label_name") == cls_a)
        cls_b_df = results_df.filter(pl.col("label_name") == cls_b)

        (cls_a_prob_a_counts, cls_a_prob_a_bins) = hist(cls_a_df[str(self.results.label_names.index(cls_a))])
        (cls_a_prob_b_counts, cls_a_prob_b_bins) = hist(cls_b_df[str(self.results.label_names.index(cls_a))])
        plt.subplot(2, 1, 1)
        plt.stairs(
            cls_a_prob_a_counts,
            cls_a_prob_a_bins,
            fill=True,
            alpha=0.4,
            label=f"{cls_a} prob. on {cls_a} label",
        )
        plt.stairs(
            cls_a_prob_b_counts,
            cls_a_prob_b_bins,
            fill=True,
            alpha=0.4,
            label=f"{cls_a} prob. on {cls_b} label",
        )
        plt.legend(loc="upper center")

        (cls_b_prob_a_counts, cls_b_prob_a_bins) = hist(cls_a_df[str(self.results.label_names.index(cls_b))])
        (cls_b_prob_b_counts, cls_b_prob_b_bins) = hist(cls_b_df[str(self.results.label_names.index(cls_b))])
        plt.subplot(2, 1, 2)
        plt.stairs(
            cls_b_prob_b_counts,
            cls_b_prob_b_bins,
            fill=True,
            alpha=0.4,
            label=f"{cls_b} prob. on {cls_b} label",
        )
        plt.stairs(
            cls_b_prob_a_counts,
            cls_b_prob_a_bins,
            fill=True,
            alpha=0.4,
            label=f"{cls_b} prob. on {cls_a} label",
        )

        plt.legend(loc="upper center")
        plt.tight_layout()
        plt.show()



"""
ViT, adapted from
https://github.com/pytorch/vision/blob/main/torchvision/models/vision_transformer.py
and
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/vision_transformer.py

Paper "An Image is Worth 16x16 Words: Transformers for Image Recognition at Scale",
https://arxiv.org/abs/2010.11929
and
Paper "Vision Transformers Need Registers", https://arxiv.org/abs/2309.16588
and
Paper "Getting ViT in Shape: Scaling Laws for Compute-Optimal Model Design", https://arxiv.org/abs/2305.13035
and
Paper "Scaling Vision Transformers", https://arxiv.org/abs/2106.04560
"""

# Reference license: BSD 3-Clause and Apache-2.0

import math
from collections.abc import Callable
from functools import partial
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.layers import FFN
from birder.layers import LayerScale
from birder.layers import MultiHeadAttentionPool
from birder.layers import SwiGLU_FFN
from birder.layers.activations import get_activation_module
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenOmissionMixin
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenOmissionResultType
from birder.net.base import TokenRetentionResultType


def adjust_position_embedding(
    pos_embedding: torch.Tensor,
    old_base_size: tuple[int, int],
    new_base_size: tuple[int, int],
    num_prefix_tokens: int,
    antialias: bool = True,
) -> torch.Tensor:
    """
    Adapted from
    https://github.com/huggingface/pytorch-image-models/blob/main/timm/layers/pos_embed.py
    """

    pos_embedding_prefix = pos_embedding[:, :num_prefix_tokens]
    pos_embedding = pos_embedding[:, num_prefix_tokens:]

    # Interpolation
    embed_dim = pos_embedding.shape[-1]
    orig_dtype = pos_embedding.dtype
    pos_embedding = pos_embedding.float()  # Interpolate needs float32
    pos_embedding = pos_embedding.reshape(1, old_base_size[0], old_base_size[1], -1).permute(0, 3, 1, 2)
    pos_embedding = F.interpolate(pos_embedding, size=new_base_size, mode="bicubic", antialias=antialias)
    pos_embedding = pos_embedding.permute(0, 2, 3, 1).reshape(1, -1, embed_dim)
    pos_embedding = pos_embedding.to(orig_dtype)

    # Add back special tokens
    return torch.concat([pos_embedding_prefix, pos_embedding], dim=1)


class PatchEmbed(nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        The entire forward is equivalent to x.flatten(2).transpose(1, 2)
        """

        (n, hidden_dim, h, w) = x.size()

        # (n, hidden_dim, h, w) -> (n, hidden_dim, (h * w))
        x = x.reshape(n, hidden_dim, h * w)

        # (n, hidden_dim, (h * w)) -> (n, (h * w), hidden_dim)
        # The self attention layer expects inputs in the format (N, S, E)
        # where S is the source sequence length, N is the batch size, E is the
        # embedding dimension
        x = x.permute(0, 2, 1)

        return x


class EncoderBlock(nn.Module):
    def __init__(
        self,
        num_heads: int,
        hidden_dim: int,
        mlp_dim: Optional[int],
        dropout: float,
        attention_dropout: float,
        drop_path: float,
        activation_layer: Callable[..., nn.Module],
        layer_scale_init_value: Optional[float] = None,
        norm_layer: Callable[..., nn.Module] = nn.LayerNorm,
        norm_layer_eps: float = 1e-6,
        mlp_layer: Callable[..., nn.Module] = FFN,
    ) -> None:
        super().__init__()
        self.need_attn = False

        if mlp_dim is None:
            mlp_dim = hidden_dim * 4

        # Attention block
        self.ln1 = norm_layer(hidden_dim, eps=norm_layer_eps)
        self.self_attention = nn.MultiheadAttention(hidden_dim, num_heads, dropout=attention_dropout, batch_first=True)
        self.drop_path1 = StochasticDepth(drop_path, mode="row")
        if layer_scale_init_value is not None:
            self.layer_scale_1 = LayerScale(hidden_dim, layer_scale_init_value)
        else:
            self.layer_scale_1 = nn.Identity()

        # MLP block
        self.ln2 = norm_layer(hidden_dim, eps=norm_layer_eps)
        self.mlp = mlp_layer(hidden_dim, mlp_dim, act_layer=activation_layer, dropout=dropout)
        self.drop_path2 = StochasticDepth(drop_path, mode="row")
        if layer_scale_init_value is not None:
            self.layer_scale_2 = LayerScale(hidden_dim, layer_scale_init_value)
        else:
            self.layer_scale_2 = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # torch._assert(x.dim() == 3, f"Expected (batch_size, seq_length, hidden_dim) got {x.size()}")
        branch1 = self.ln1(x)
        (branch1, _) = self.self_attention(
            branch1, branch1, branch1, need_weights=self.need_attn, average_attn_weights=False
        )
        branch1 = self.layer_scale_1(branch1)
        branch1 = self.drop_path1(branch1) + x

        branch2 = self.ln2(branch1)
        branch2 = self.mlp(branch2)
        branch2 = self.layer_scale_2(branch2)

        x = self.drop_path2(branch2) + branch1

        return x

    def set_need_attn(self) -> None:
        self.need_attn = True


class Encoder(nn.Module):
    def __init__(
        self,
        num_layers: int,
        num_heads: int,
        hidden_dim: int,
        mlp_dim: int,
        dropout: float,
        attention_dropout: float,
        dpr: list[float],
        pre_norm: bool = False,
        activation_layer: Callable[..., nn.Module] = nn.GELU,
        layer_scale_init_value: Optional[float] = None,
        norm_layer: Callable[..., nn.Module] = nn.LayerNorm,
        norm_layer_eps: float = 1e-6,
        mlp_layer: Callable[..., nn.Module] = FFN,
    ) -> None:
        super().__init__()
        pre_layers = []
        if dropout > 0.0:
            pre_layers.append(nn.Dropout(dropout))
        if pre_norm is True:
            pre_layers.append(norm_layer(hidden_dim, eps=norm_layer_eps))

        self.pre_block = nn.Sequential(*pre_layers)

        layers = []
        for i in range(num_layers):
            layers.append(
                EncoderBlock(
                    num_heads,
                    hidden_dim,
                    mlp_dim,
                    dropout,
                    attention_dropout,
                    dpr[i],
                    activation_layer=activation_layer,
                    layer_scale_init_value=layer_scale_init_value,
                    norm_layer=norm_layer,
                    norm_layer_eps=norm_layer_eps,
                    mlp_layer=mlp_layer,
                )
            )

        self.block = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.pre_block(x)
        return self.block(x)

    def forward_features(self, x: torch.Tensor) -> list[torch.Tensor]:
        x = self.pre_block(x)

        xs = []
        for blk in self.block:
            x = blk(x)
            xs.append(x)

        return xs

    def set_need_attn(self) -> None:
        for b in self.block:
            b.set_need_attn()


# pylint: disable=too-many-instance-attributes
class ViT(DetectorBackbone, PreTrainEncoder, MaskedTokenOmissionMixin, MaskedTokenRetentionMixin):
    block_group_regex = r"encoder\.block\.(\d+)"

    # pylint: disable=too-many-locals,too-many-branches
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        image_size = self.size
        attention_dropout = 0.0
        dropout = 0.0
        pos_embed_special_tokens: bool = self.config.get("pos_embed_special_tokens", True)
        patch_size: int = self.config["patch_size"]
        num_layers: int = self.config["num_layers"]
        num_heads: int = self.config["num_heads"]
        hidden_dim: int = self.config["hidden_dim"]
        mlp_dim: int = self.config["mlp_dim"]
        layer_scale_init_value: Optional[float] = self.config.get("layer_scale_init_value", None)
        pre_norm: bool = self.config.get("pre_norm", False)
        post_norm: bool = self.config.get("post_norm", True)
        num_reg_tokens: int = self.config.get("num_reg_tokens", 0)
        class_token: bool = self.config.get("class_token", True)
        attn_pool_head: bool = self.config.get("attn_pool_head", False)
        attn_pool_num_heads: Optional[int] = self.config.get("attn_pool_num_heads", None)
        attn_pool_special_tokens: bool = self.config.get("attn_pool_special_tokens", False)
        norm_layer_type: str = self.config.get("norm_layer_type", "LayerNorm")
        norm_layer_eps: float = self.config.get("norm_layer_eps", 1e-6)
        mlp_layer_type: str = self.config.get("mlp_layer_type", "FFN")
        act_layer_type: Optional[str] = self.config.get("act_layer_type", None)  # Default according to mlp type
        drop_path_rate: float = self.config["drop_path_rate"]

        if norm_layer_type == "LayerNorm":
            norm_layer = nn.LayerNorm
        elif norm_layer_type == "RMSNorm":
            norm_layer = nn.RMSNorm
        else:
            raise ValueError(f"Unknown norm_layer_type '{norm_layer_type}'")

        if mlp_layer_type == "FFN":
            mlp_layer = FFN
            act_layer = nn.GELU
        elif mlp_layer_type == "SwiGLU_FFN":
            mlp_layer = SwiGLU_FFN
            act_layer = nn.SiLU
        else:
            raise ValueError(f"Unknown mlp_layer_type '{mlp_layer_type}'")

        if act_layer_type is not None:
            act_layer = get_activation_module(act_layer_type)

        torch._assert(image_size[0] % patch_size == 0, "Input shape indivisible by patch size!")
        torch._assert(image_size[1] % patch_size == 0, "Input shape indivisible by patch size!")
        torch._assert(hidden_dim % num_heads == 0, "Hidden dim indivisible by num heads!")
        self.pos_embed_special_tokens = pos_embed_special_tokens
        self.patch_size = patch_size
        self.num_layers = num_layers
        self.hidden_dim = hidden_dim
        self.num_reg_tokens = num_reg_tokens
        self.attn_pool_special_tokens = attn_pool_special_tokens
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, num_layers)]  # Stochastic depth decay rule

        self.conv_proj = nn.Conv2d(
            self.input_channels,
            hidden_dim,
            kernel_size=(patch_size, patch_size),
            stride=(patch_size, patch_size),
            padding=(0, 0),
            bias=not pre_norm,
        )
        self.patch_embed = PatchEmbed()

        seq_length = (image_size[0] // patch_size) * (image_size[1] // patch_size)
        self.num_special_tokens = 0

        # Add a class token
        if class_token is True:
            self.class_token = nn.Parameter(torch.zeros(1, 1, hidden_dim))
            self.num_special_tokens += 1
            if pos_embed_special_tokens is True:
                seq_length += 1
        else:
            self.class_token = None

        # Add optional register tokens
        if self.num_reg_tokens > 0:
            self.reg_tokens = nn.Parameter(torch.zeros(1, self.num_reg_tokens, hidden_dim))
            self.num_special_tokens += self.num_reg_tokens
            if pos_embed_special_tokens is True:
                seq_length += self.num_reg_tokens
        else:
            self.reg_tokens = None

        # Add positional embedding
        self.pos_embedding = nn.Parameter(torch.empty(1, seq_length, hidden_dim).normal_(std=0.02))  # from BERT

        self.encoder = Encoder(
            num_layers,
            num_heads,
            hidden_dim,
            mlp_dim,
            dropout,
            attention_dropout,
            dpr,
            pre_norm=pre_norm,
            activation_layer=act_layer,
            layer_scale_init_value=layer_scale_init_value,
            norm_layer=norm_layer,
            norm_layer_eps=norm_layer_eps,
            mlp_layer=mlp_layer,
        )

        if post_norm is True:
            self.norm = norm_layer(hidden_dim, eps=norm_layer_eps)
        else:
            self.norm = nn.Identity()

        if attn_pool_head is False:
            self.attn_pool = None
        else:
            if attn_pool_num_heads is None:
                attn_pool_num_heads = num_heads

            self.attn_pool = MultiHeadAttentionPool(hidden_dim, attn_pool_num_heads, mlp_dim, qkv_bias=True)

        self.return_stages = ["neck"]  # Actually meaningless, just for completeness
        self.return_channels = [hidden_dim]
        self.embedding_size = hidden_dim
        self.classifier = self.create_classifier()

        self.max_stride = patch_size
        self.stem_stride = patch_size
        self.stem_width = hidden_dim
        self.encoding_size = hidden_dim
        self.decoder_block = partial(
            EncoderBlock,
            16,
            mlp_dim=None,
            dropout=0,
            attention_dropout=0,
            drop_path=0,
            activation_layer=act_layer,
            norm_layer=norm_layer,
            mlp_layer=mlp_layer,
        )

        # Weight initialization
        if isinstance(self.conv_proj, nn.Conv2d):
            # Init the patchify stem
            fan_in = self.conv_proj.in_channels * self.conv_proj.kernel_size[0] * self.conv_proj.kernel_size[1]
            nn.init.trunc_normal_(self.conv_proj.weight, std=math.sqrt(1 / fan_in))
            if self.conv_proj.bias is not None:
                nn.init.zeros_(self.conv_proj.bias)

        if isinstance(self.classifier, nn.Linear):
            nn.init.zeros_(self.classifier.weight)
            nn.init.zeros_(self.classifier.bias)

    def _get_pos_embed(self, H: int, W: int) -> torch.Tensor:
        if self.dynamic_size is False:
            return self.pos_embedding

        if H == self.size[0] and W == self.size[1]:
            return self.pos_embedding

        return adjust_position_embedding(
            self.pos_embedding,
            (self.size[0] // self.patch_size, self.size[1] // self.patch_size),
            (H // self.patch_size, W // self.patch_size),
            self.num_special_tokens if self.pos_embed_special_tokens is True else 0,
            antialias=False,
        )

    def freeze(self, freeze_classifier: bool = True, unfreeze_features: bool = False) -> None:
        for param in self.parameters():
            param.requires_grad = False

        if freeze_classifier is False:
            for param in self.classifier.parameters():
                param.requires_grad = True

        if unfreeze_features is True:
            if self.attn_pool is not None:
                for param in self.attn_pool.parameters():
                    param.requires_grad = True

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        (H, W) = x.shape[-2:]
        x = self.conv_proj(x)
        x = self.patch_embed(x)

        if self.pos_embed_special_tokens is False:
            x = x + self._get_pos_embed(H, W)

        # Expand the class token to the full batch
        if self.class_token is not None:
            batch_class_token = self.class_token.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_class_token, x], dim=1)

        # Expand the register tokens to the full batch
        if self.reg_tokens is not None:
            batch_reg_tokens = self.reg_tokens.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_reg_tokens, x], dim=1)

        if self.pos_embed_special_tokens is True:
            x = x + self._get_pos_embed(H, W)

        x = self.encoder(x)
        x = self.norm(x)

        x = x[:, self.num_special_tokens :]
        x = x.permute(0, 2, 1)
        (B, C, _) = x.size()
        x = x.reshape(B, C, H // self.patch_size, W // self.patch_size)

        return {self.return_stages[0]: x}

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.conv_proj.parameters():
            param.requires_grad = False

        self.pos_embedding.requires_grad = False

        for idx, module in enumerate(self.encoder.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    # pylint: disable=too-many-branches
    def masked_encoding_omission(
        self,
        x: torch.Tensor,
        ids_keep: Optional[torch.Tensor] = None,
        return_all_features: bool = False,
        return_keys: Literal["all", "tokens", "embedding"] = "tokens",
    ) -> TokenOmissionResultType:
        (H, W) = x.shape[-2:]

        # Reshape and permute the input tensor
        x = self.conv_proj(x)
        x = self.patch_embed(x)

        # Add pos embedding without special tokens
        pos_embedding = self._get_pos_embed(H, W)
        if self.pos_embed_special_tokens is True:
            x = x + pos_embedding[:, self.num_special_tokens :, :]
        else:
            x = x + pos_embedding

        # Mask tokens
        if ids_keep is not None:
            x = torch.gather(x, dim=1, index=ids_keep.unsqueeze(-1).repeat(1, 1, x.size(2)))

        # Append class and register tokens
        if self.class_token is not None:
            if self.pos_embed_special_tokens is True:
                cls_token = self.class_token + pos_embedding[:, self.num_reg_tokens : self.num_reg_tokens + 1, :]
            else:
                cls_token = self.class_token

            batch_class_token = cls_token.expand(x.shape[0], -1, -1)
            x = torch.concat((batch_class_token, x), dim=1)

        if self.reg_tokens is not None:
            if self.pos_embed_special_tokens is True:
                reg_tokens = self.reg_tokens + pos_embedding[:, 0 : self.num_reg_tokens, :]
            else:
                reg_tokens = self.reg_tokens

            batch_reg_tokens = reg_tokens.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_reg_tokens, x], dim=1)

        # Apply transformer
        if return_all_features is True:
            xs = self.encoder.forward_features(x)
            xs[-1] = self.norm(xs[-1])
            x = torch.stack(xs, dim=-1)
        else:
            x = self.encoder(x)
            x = self.norm(x)

        result: TokenOmissionResultType = {}
        if return_keys in ("all", "tokens"):
            result["tokens"] = x

        if return_keys in ("all", "embedding"):
            if return_all_features is True:
                x = x[..., -1]

            if self.attn_pool is not None:
                if self.attn_pool_special_tokens is False:
                    x = x[:, self.num_special_tokens :]

                x = self.attn_pool(x)
                result["embedding"] = x[:, 0]
            elif self.class_token is None:
                x = x[:, self.num_special_tokens :]
                result["embedding"] = x.mean(dim=1)
            else:
                result["embedding"] = x[:, self.num_reg_tokens]

        return result

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        (H, W) = x.shape[-2:]

        x = self.conv_proj(x)
        x = mask_tensor(x, mask, mask_token=mask_token, patch_factor=self.max_stride // self.stem_stride)

        # Reshape and permute the input tensor
        x = self.patch_embed(x)

        if self.pos_embed_special_tokens is False:
            x = x + self._get_pos_embed(H, W)

        # Expand the class token to the full batch
        if self.class_token is not None:
            batch_class_token = self.class_token.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_class_token, x], dim=1)

        # Expand the register tokens to the full batch
        if self.reg_tokens is not None:
            batch_reg_tokens = self.reg_tokens.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_reg_tokens, x], dim=1)

        if self.pos_embed_special_tokens is True:
            x = x + self._get_pos_embed(H, W)

        x = self.encoder(x)
        x = self.norm(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            features = x[:, self.num_special_tokens :]
            features = features.permute(0, 2, 1)
            (B, C, _) = features.size()
            features = features.reshape(B, C, H // self.patch_size, W // self.patch_size)
            result["features"] = features

        if return_keys in ("all", "embedding"):
            if self.attn_pool is not None:
                if self.attn_pool_special_tokens is False:
                    x = x[:, self.num_special_tokens :]

                x = self.attn_pool(x)
                result["embedding"] = x[:, 0]
            elif self.class_token is None:
                x = x[:, self.num_special_tokens :]
                result["embedding"] = x.mean(dim=1)
            else:
                result["embedding"] = x[:, self.num_reg_tokens]

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        (H, W) = x.shape[-2:]

        # Reshape and permute the input tensor
        x = self.conv_proj(x)
        x = self.patch_embed(x)

        if self.pos_embed_special_tokens is False:
            x = x + self._get_pos_embed(H, W)

        # Expand the class token to the full batch
        if self.class_token is not None:
            batch_class_token = self.class_token.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_class_token, x], dim=1)

        # Expand the register tokens to the full batch
        if self.reg_tokens is not None:
            batch_reg_tokens = self.reg_tokens.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_reg_tokens, x], dim=1)

        if self.pos_embed_special_tokens is True:
            x = x + self._get_pos_embed(H, W)

        x = self.encoder(x)
        x = self.norm(x)

        return x

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)

        if self.attn_pool is not None:
            if self.attn_pool_special_tokens is False:
                x = x[:, self.num_special_tokens :]

            x = self.attn_pool(x)
            return x[:, 0]

        if self.class_token is None:
            x = x[:, self.num_special_tokens :]
            return x.mean(dim=1)

        # Classifier "token" as used by standard language architectures
        return x[:, self.num_reg_tokens]

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        assert new_size[0] % self.patch_size == 0, "Input shape indivisible by patch size!"
        assert new_size[1] % self.patch_size == 0, "Input shape indivisible by patch size!"

        old_size = self.size
        super().adjust_size(new_size)

        if self.pos_embed_special_tokens is True:
            num_prefix_tokens = self.num_special_tokens
        else:
            num_prefix_tokens = 0

        # Add back class tokens
        self.pos_embedding = nn.Parameter(
            # On rounding error see: https://github.com/facebookresearch/dino/issues/8
            adjust_position_embedding(
                self.pos_embedding,
                (old_size[0] // self.patch_size, old_size[1] // self.patch_size),
                (new_size[0] // self.patch_size, new_size[1] // self.patch_size),
                num_prefix_tokens,
            )
        )


# For the model naming convention see rope_vit.py

registry.register_model_config(
    "vit_t32",
    ViT,
    config={
        "patch_size": 32,
        "num_layers": 12,
        "num_heads": 3,
        "hidden_dim": 192,
        "mlp_dim": 768,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "vit_t16",
    ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 3,
        "hidden_dim": 192,
        "mlp_dim": 768,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "vit_s32",
    ViT,
    config={
        "patch_size": 32,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "vit_s16",
    ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "vit_s16_pn",
    ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "pre_norm": True,
        "norm_layer_eps": 1e-5,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "vit_s14",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "vit_m32",
    ViT,
    config={
        "patch_size": 32,
        "num_layers": 12,
        "num_heads": 8,
        "hidden_dim": 512,
        "mlp_dim": 2048,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "vit_m16",
    ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 8,
        "hidden_dim": 512,
        "mlp_dim": 2048,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "vit_m14",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 12,
        "num_heads": 8,
        "hidden_dim": 512,
        "mlp_dim": 2048,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "vit_b32",
    ViT,
    config={
        "patch_size": 32,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "vit_b16",
    ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_b16_pn_quick_gelu",
    ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "pre_norm": True,
        "norm_layer_eps": 1e-5,
        "act_layer_type": "quick_gelu",
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_b14",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_l32",
    ViT,
    config={
        "patch_size": 32,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_l16",
    ViT,
    config={
        "patch_size": 16,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_l14",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_l14_pn",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "pre_norm": True,
        "norm_layer_eps": 1e-5,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_l14_pn_quick_gelu",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "pre_norm": True,
        "norm_layer_eps": 1e-5,
        "act_layer_type": "quick_gelu",
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_h16",
    ViT,
    config={
        "patch_size": 16,
        "num_layers": 32,
        "num_heads": 16,
        "hidden_dim": 1280,
        "mlp_dim": 5120,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_h14",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 32,
        "num_heads": 16,
        "hidden_dim": 1280,
        "mlp_dim": 5120,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(  # From "Scaling Vision Transformers"
    "vit_g14",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 40,
        "num_heads": 16,
        "hidden_dim": 1408,
        "mlp_dim": 6144,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(  # From "Scaling Vision Transformers"
    "vit_gigantic14",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 48,
        "num_heads": 16,
        "hidden_dim": 1664,
        "mlp_dim": 8192,
        "drop_path_rate": 0.1,
    },
)

# With registers
registry.register_model_config(
    "vit_reg1_t16",
    ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 3,
        "hidden_dim": 192,
        "mlp_dim": 768,
        "num_reg_tokens": 1,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "vit_reg1_s32",
    ViT,
    config={
        "patch_size": 32,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "num_reg_tokens": 1,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "vit_reg1_s16",
    ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "num_reg_tokens": 1,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "vit_reg1_s16_ls",
    ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "layer_scale_init_value": 1e-5,
        "num_reg_tokens": 1,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "vit_reg1_s16_rms_ls",
    ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "layer_scale_init_value": 1e-5,
        "num_reg_tokens": 1,
        "norm_layer_type": "RMSNorm",
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "vit_reg1_s14",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 12,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "num_reg_tokens": 1,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "vit_reg4_m32",
    ViT,
    config={
        "patch_size": 32,
        "num_layers": 12,
        "num_heads": 8,
        "hidden_dim": 512,
        "mlp_dim": 2048,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "vit_reg4_m16",
    ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 8,
        "hidden_dim": 512,
        "mlp_dim": 2048,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "vit_reg4_m16_rms_avg",
    ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 8,
        "hidden_dim": 512,
        "mlp_dim": 2048,
        "num_reg_tokens": 4,
        "class_token": False,
        "norm_layer_type": "RMSNorm",
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "vit_reg4_m14",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 12,
        "num_heads": 8,
        "hidden_dim": 512,
        "mlp_dim": 2048,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "vit_reg4_b32",
    ViT,
    config={
        "patch_size": 32,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "vit_reg4_b16",
    ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_reg4_b16_avg",
    ViT,
    config={
        "patch_size": 16,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "num_reg_tokens": 4,
        "class_token": False,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_reg4_b14",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_reg8_b14_ap",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 12,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "num_reg_tokens": 8,
        "class_token": False,
        "attn_pool_head": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_reg4_l32",
    ViT,
    config={
        "patch_size": 32,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_reg4_l16",
    ViT,
    config={
        "patch_size": 16,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_reg4_l14",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(  # DeiT III style
    "vit_reg4_l14_nps_ls",
    ViT,
    config={
        "pos_embed_special_tokens": False,
        "patch_size": 14,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "layer_scale_init_value": 1e-5,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_reg8_l14_ap",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "num_reg_tokens": 8,
        "class_token": False,
        "attn_pool_head": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_reg8_l14_rms_ap",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 24,
        "num_heads": 16,
        "hidden_dim": 1024,
        "mlp_dim": 4096,
        "num_reg_tokens": 8,
        "class_token": False,
        "attn_pool_head": True,
        "norm_layer_type": "RMSNorm",
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_reg4_h16",
    ViT,
    config={
        "patch_size": 16,
        "num_layers": 32,
        "num_heads": 16,
        "hidden_dim": 1280,
        "mlp_dim": 5120,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_reg4_h14",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 32,
        "num_heads": 16,
        "hidden_dim": 1280,
        "mlp_dim": 5120,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(  # From "Scaling Vision Transformers"
    "vit_reg4_g14",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 40,
        "num_heads": 16,
        "hidden_dim": 1408,
        "mlp_dim": 6144,
        "num_reg_tokens": 4,
        "drop_path_rate": 0.1,
    },
)

# Shape-optimized vision transformer (SoViT)
registry.register_model_config(
    "vit_so150m_p14_ap",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 18,
        "num_heads": 16,
        "hidden_dim": 896,  # Changed from 880 for RoPE divisibility
        "mlp_dim": 2320,
        "class_token": False,
        "attn_pool_head": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_so400m_p14_ap",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 27,
        "num_heads": 16,
        "hidden_dim": 1152,
        "mlp_dim": 4304,
        "class_token": False,
        "attn_pool_head": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_reg8_so150m_p16_swiglu_ap",
    ViT,
    config={
        "patch_size": 16,
        "num_layers": 18,
        "num_heads": 16,
        "hidden_dim": 896,  # Changed from 880 for RoPE divisibility
        "mlp_dim": 2320,
        "num_reg_tokens": 8,
        "class_token": False,
        "attn_pool_head": True,
        "mlp_layer_type": "SwiGLU_FFN",
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_reg4_so150m_p14_ap",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 18,
        "num_heads": 16,
        "hidden_dim": 896,  # Changed from 880 for RoPE divisibility
        "mlp_dim": 2320,
        "num_reg_tokens": 4,
        "class_token": False,
        "attn_pool_head": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_reg8_so150m_p14_avg",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 18,
        "num_heads": 16,
        "hidden_dim": 896,  # Changed from 880 for RoPE divisibility
        "mlp_dim": 2320,
        "num_reg_tokens": 8,
        "class_token": False,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_reg8_so150m_p14_swiglu",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 18,
        "num_heads": 16,
        "hidden_dim": 896,  # Changed from 880 for RoPE divisibility
        "mlp_dim": 2320,
        "num_reg_tokens": 8,
        "mlp_layer_type": "SwiGLU_FFN",
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_reg8_so150m_p14_swiglu_avg",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 18,
        "num_heads": 16,
        "hidden_dim": 896,  # Changed from 880 for RoPE divisibility
        "mlp_dim": 2320,
        "num_reg_tokens": 8,
        "class_token": False,
        "mlp_layer_type": "SwiGLU_FFN",
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_reg8_so150m_p14_ap",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 18,
        "num_heads": 16,
        "hidden_dim": 896,  # Changed from 880 for RoPE divisibility
        "mlp_dim": 2320,
        "num_reg_tokens": 8,
        "class_token": False,
        "attn_pool_head": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_reg4_so400m_p14_ap",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 27,
        "num_heads": 16,
        "hidden_dim": 1152,
        "mlp_dim": 4304,
        "num_reg_tokens": 4,
        "class_token": False,
        "attn_pool_head": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "vit_reg8_so400m_p14_ap",
    ViT,
    config={
        "patch_size": 14,
        "num_layers": 27,
        "num_heads": 16,
        "hidden_dim": 1152,
        "mlp_dim": 4304,
        "num_reg_tokens": 8,
        "class_token": False,
        "attn_pool_head": True,
        "drop_path_rate": 0.1,
    },
)

registry.register_weights(
    "vit_l16_mim_200",
    {
        "url": "https://huggingface.co/birder-project/vit_l16_mim/resolve/main",
        "description": (
            "ViT l16 image encoder pre-trained using Masked Image Modeling (MIM) for 200 epochs. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 1157.1,
                "sha256": "003b15a79cd528339de1b19304bbd04fd5885df36b80e19202cd6ef6f8ffbed1",
            },
        },
        "net": {"network": "vit_l16", "tag": "mim"},
    },
)
registry.register_weights(
    "vit_l16_mim_400",
    {
        "url": "https://huggingface.co/birder-project/vit_l16_mim/resolve/main",
        "description": (
            "ViT l16 image encoder pre-trained using Masked Image Modeling (MIM) for 400 epochs. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 1157.1,
                "sha256": "c6083c6532996addaf4efe29276aa55f9a3c77984f862f720c6131f86b847994",
            },
        },
        "net": {"network": "vit_l16", "tag": "mim"},
    },
)
registry.register_weights(
    "vit_l16_mim-eu-common",
    {
        "url": "https://huggingface.co/birder-project/vit_l16_mim-eu-common/resolve/main",
        "description": "ViT l16 model with MIM pretraining, then fine-tuned on the eu-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 1160.1,
                "sha256": "3b7235b90f76fb1e0e36d4c4111777a4cc4e4500552fe840c51170b208310d16",
            },
        },
        "net": {"network": "vit_l16", "tag": "mim-eu-common"},
    },
)
registry.register_weights(  # BioCLIP v2: https://arxiv.org/abs/2505.23883
    "vit_l14_pn_bioclip-v2",
    {
        "url": "https://huggingface.co/birder-project/vit_l14_pn_bioclip-v2/resolve/main",
        "description": (
            "ViT l14 image encoder pre-trained by Imageomics using CLIP on the TreeOfLife-200M dataset. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 1156.6,
                "sha256": "cfb998d762cd2ba883964026ddfc8f2f84cf1e6ad6f7264ab33da52f57d25fab",
            },
        },
        "net": {"network": "vit_l14_pn", "tag": "bioclip-v2"},
    },
)
registry.register_weights(  # OpenAI CLIP: https://arxiv.org/abs/2103.00020
    "vit_l14_pn_quick_gelu_openai-clip",
    {
        "url": "https://huggingface.co/birder-project/vit_l14_pn_quick_gelu_openai-clip/resolve/main",
        "description": (
            "ViT l14 image encoder pre-trained by OpenAI using CLIP. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 1159.7,
                "sha256": "e4c6ff7467608c412d35f9a4e2df18f3b8f05fc9eca3803c8fcc01558921378d",
            },
        },
        "net": {"network": "vit_l14_pn_quick_gelu", "tag": "openai-clip"},
    },
)
registry.register_weights(  # SigLIP 2: https://arxiv.org/abs/2502.14786
    "vit_so400m_p14_ap_siglip-v2-webli",
    {
        "url": "https://huggingface.co/birder-project/vit_so400m_p14_ap_siglip-v2-webli/resolve/main",
        "description": (
            "ViT SO400m image encoder pre-trained by Google using SigLIP. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 1631.6,
                "sha256": "1f9f659a7b1bdf8a6a2977140be9bb3f876f7f756bf6e13d54bf00f3b6db0b0f",
            },
        },
        "net": {"network": "vit_so400m_p14_ap", "tag": "siglip-v2-webli"},
    },
)

# With registers
registry.register_weights(
    "vit_reg4_m16_rms_avg_i-jepa",
    {
        "url": "https://huggingface.co/birder-project/vit_reg4_m16_rms_avg_i-jepa/resolve/main",
        "description": (
            "ViT reg4 m16 RMS norm image encoder pre-trained using I-JEPA"
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 146.2,
                "sha256": "bc4c9e600e93322440fb68c1001216d49c54c7587cdf61544f363f9537152f4a",
            },
        },
        "net": {"network": "vit_reg4_m16_rms_avg", "tag": "i-jepa"},
    },
)
registry.register_weights(
    "vit_reg4_m16_rms_avg_i-jepa-inat21-256px",
    {
        "url": "https://huggingface.co/birder-project/vit_reg4_m16_rms_avg_i-jepa-inat21/resolve/main",
        "description": (
            "ViT reg4 m16 RMS norm image encoder pre-trained using I-JEPA"
            "then fine-tuned on the iNaturalist 2021 dataset"
        ),
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 166.8,
                "sha256": "9ff659be9826bbbafbcfa85d79d0fa9d5ac383fd2442ffa36db6c4f7ab09b86a",
            },
        },
        "net": {"network": "vit_reg4_m16_rms_avg", "tag": "i-jepa-inat21-256px"},
    },
)
registry.register_weights(
    "vit_reg4_m16_rms_avg_i-jepa-inat21",
    {
        "url": "https://huggingface.co/birder-project/vit_reg4_m16_rms_avg_i-jepa-inat21/resolve/main",
        "description": (
            "ViT reg4 m16 RMS norm image encoder pre-trained using I-JEPA"
            "then fine-tuned on the iNaturalist 2021 dataset"
        ),
        "resolution": (384, 384),
        "formats": {
            "pt": {
                "file_size": 167.4,
                "sha256": "1cfa7ebea3db95363bf9e35fc24be94e419debe5db58746fe3320fbcb8bda6dd",
            },
        },
        "net": {"network": "vit_reg4_m16_rms_avg", "tag": "i-jepa-inat21"},
    },
)
registry.register_weights(
    "vit_reg4_m16_rms_avg_i-jepa-imagenet21k",
    {
        "url": "https://huggingface.co/birder-project/vit_reg4_m16_rms_avg_i-jepa-imagenet21k/resolve/main",
        "description": (
            "ViT reg4 m16 RMS norm image encoder pre-trained using I-JEPA then fine-tuned on the ImageNet-21K dataset"
        ),
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 184.2,
                "sha256": "d6d9fc47ecbad04a83b178bcd2eeecbd77569cc2a17fbdf52e02feda54523c3f",
            },
        },
        "net": {"network": "vit_reg4_m16_rms_avg", "tag": "i-jepa-imagenet21k"},
    },
)
registry.register_weights(
    "vit_reg4_b16_mim_200",
    {
        "url": "https://huggingface.co/birder-project/vit_reg4_b16_mim/resolve/main",
        "description": (
            "ViT reg4 b16 image encoder pre-trained using Masked Image Modeling (MIM) for 200 epochs. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 327.4,
                "sha256": "6b044cd7834293e344309f809070db3fe9ede489478e7549ad96255f9d76b329",
            },
        },
        "net": {"network": "vit_reg4_b16", "tag": "mim"},
    },
)
registry.register_weights(
    "vit_reg4_b16_mim_300",
    {
        "url": "https://huggingface.co/birder-project/vit_reg4_b16_mim/resolve/main",
        "description": (
            "ViT reg4 b16 image encoder pre-trained using Masked Image Modeling (MIM) for 300 epochs. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 327.4,
                "sha256": "e0df2e79f8ed0612d12c736cc6317be1b9b354e468715a5077366f7676fdd2ce",
            },
        },
        "net": {"network": "vit_reg4_b16", "tag": "mim"},
    },
)
registry.register_weights(
    "vit_reg4_b16_mim-intermediate-il-common",
    {
        "url": "https://huggingface.co/birder-project/vit_reg4_b16_mim-intermediate-il-common/resolve/main",
        "description": (
            "ViT reg4 b16 model with MIM pretraining and intermediate training, "
            "then fine-tuned on the il-common dataset"
        ),
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 328.7,
                "sha256": "3d1564be46b23081c76aa87c7e90324214b6ced899d4b38d59d1a4154b13f01c",
            },
        },
        "net": {"network": "vit_reg4_b16", "tag": "mim-intermediate-il-common"},
    },
)
registry.register_weights(
    "vit_reg4_b16_mim-intermediate-arabian-peninsula",
    {
        "url": "https://huggingface.co/birder-project/vit_reg4_b16_mim-intermediate-arabian-peninsula/resolve/main",
        "description": (
            "ViT reg4 b16 model with MIM pretraining and intermediate training, "
            "then fine-tuned on the arabian-peninsula dataset"
        ),
        "resolution": (384, 384),
        "formats": {
            "pt": {
                "file_size": 330.7,
                "sha256": "e011f931a5a4d96ef21283d70911a55ea649eadfefa9c163a48b996797f0d9da",
            },
        },
        "net": {"network": "vit_reg4_b16", "tag": "mim-intermediate-arabian-peninsula"},
    },
)
registry.register_weights(  # DINO v2: https://arxiv.org/abs/2304.07193
    "vit_reg4_l14_nps_ls_dino-v2-lvd142m",
    {
        "url": "https://huggingface.co/birder-project/vit_reg4_l14_nps_ls_dino-v2-lvd142m/resolve/main",
        "description": (
            "ViT reg4 l14 image encoder pre-trained by Facebook AI using DINOv2. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (518, 518),
        "formats": {
            "pt": {
                "file_size": 1161.2,
                "sha256": "56d39cbaed8b7da72175b7b3a0c9419e71aabc1e9516567703a39ba05244a44f",
            },
        },
        "net": {"network": "vit_reg4_l14_nps_ls", "tag": "dino-v2-lvd142m"},
    },
)



import typing
from collections.abc import Callable
from pathlib import Path
from typing import Any
from typing import Literal
from typing import Optional

import torch
from torchvision.datasets import ImageFolder
from torchvision.datasets.folder import default_loader

from birder.conf import settings
from birder.datahub._lib import download_url
from birder.datahub._lib import extract_archive

SplitType = Literal["training", "validation", "testing"]


class TestDataset(ImageFolder):
    """
    Name: Birder TestDataset
    """

    def __init__(
        self,
        target_dir: Optional[str | Path] = None,
        download: bool = False,
        split: SplitType = "training",
        transform: Optional[Callable[..., torch.Tensor]] = None,
        target_transform: Optional[Callable[..., Any]] = None,
        loader: Callable[[str], Any] = default_loader,
        is_valid_file: Optional[Callable[[str], bool]] = None,
        progress_bar: bool = True,
    ) -> None:
        if target_dir is None:
            target_dir = settings.DATA_DIR

        if isinstance(target_dir, str):
            target_dir = Path(target_dir)

        self._target_dir: Path = target_dir
        self._root = self._target_dir.joinpath("TestDataset")
        if download is True:
            src = self._target_dir.joinpath("TestDataset.tar")
            downloaded = download_url(
                "https://f000.backblazeb2.com/file/birder/data/TestDataset.tar",
                src,
                sha256="28ca71c6308742ad8ebb47d5bd3de4db1f26b3173fa568d00abc1df05b556916",
                progress_bar=progress_bar,
            )
            if downloaded is True or self._root.exists() is False:
                extract_archive(src, self._root)

        else:
            # Some sanity checks
            if self._root.exists() is False or self._root.is_dir() is False:
                raise RuntimeError("Dataset not found, try download=True to download it")

            for split_name in ["training", "validation"]:
                if self._root.joinpath(split_name).exists() is False:
                    raise RuntimeError("Dataset seems corrupted")

        super().__init__(self._root.joinpath(split), transform, target_transform, loader, is_valid_file)

    def __getitem__(self, index: int) -> tuple[str, torch.Tensor, Any]:
        (path, target) = self.samples[index]
        sample = self.loader(path)
        if self.transform is not None:
            sample = self.transform(sample)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return (path, sample, target)


class Flowers102(ImageFolder):
    """
    Name: 102 Flowers
    Link: https://www.robots.ox.ac.uk/~vgg/data/flowers/102/
    Size: 102 categories dataset consists of between 40 and 258 images per class (337MB)
    """

    def __init__(
        self,
        target_dir: Optional[str | Path] = None,
        download: bool = False,
        split: SplitType = "training",
        transform: Optional[Callable[..., torch.Tensor]] = None,
        target_transform: Optional[Callable[..., Any]] = None,
        loader: Callable[[str], Any] = default_loader,
        is_valid_file: Optional[Callable[[str], bool]] = None,
        progress_bar: bool = True,
    ) -> None:
        if target_dir is None:
            target_dir = settings.DATA_DIR

        if isinstance(target_dir, str):
            target_dir = Path(target_dir)

        self._target_dir: Path = target_dir
        self._root = self._target_dir.joinpath("Flowers102")
        if download is True:
            src = self._target_dir.joinpath("Flowers102.tar")
            downloaded = download_url(
                "https://f000.backblazeb2.com/file/birder/data/Flowers102.tar",
                src,
                sha256="6e7cf83821ed267e178dbabdb66cb4e23643aaf9d6180c3d93929f11cfbc4582",
                progress_bar=progress_bar,
            )
            if downloaded is True or self._root.exists() is False:
                extract_archive(src, self._root)

        else:
            # Some sanity checks
            if self._root.exists() is False or self._root.is_dir() is False:
                raise RuntimeError("Dataset not found, try download=True to download it")

            for split_name in typing.get_args(SplitType):
                if self._root.joinpath(split_name).exists() is False:
                    raise RuntimeError("Dataset seems corrupted")

        super().__init__(self._root.joinpath(split), transform, target_transform, loader, is_valid_file)

    def __getitem__(self, index: int) -> tuple[str, torch.Tensor, Any]:
        (path, target) = self.samples[index]
        sample = self.loader(path)
        if self.transform is not None:
            sample = self.transform(sample)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return (path, sample, target)


# pylint: disable=invalid-name
class CUB_200_2011(ImageFolder):
    """
    Name: CUB_200_2011
    Link: https://www.vision.caltech.edu/datasets/cub_200_2011/
    Size: 200 categories dataset consists of 11,788 images (1.1GB)
    """

    def __init__(
        self,
        target_dir: Optional[str | Path] = None,
        download: bool = False,
        split: SplitType = "training",
        transform: Optional[Callable[..., torch.Tensor]] = None,
        target_transform: Optional[Callable[..., Any]] = None,
        loader: Callable[[str], Any] = default_loader,
        is_valid_file: Optional[Callable[[str], bool]] = None,
        progress_bar: bool = True,
    ) -> None:
        if target_dir is None:
            target_dir = settings.DATA_DIR

        if isinstance(target_dir, str):
            target_dir = Path(target_dir)

        self._target_dir: Path = target_dir
        self._root = self._target_dir.joinpath("CUB_200_2011")
        if download is True:
            src = self._target_dir.joinpath("CUB_200_2011.tar")
            downloaded = download_url(
                "https://huggingface.co/datasets/birder-project/CUB_200_2011/resolve/main/CUB_200_2011.tar",
                src,
                sha256="acb58211efa4253d59935572b6d1d3b9f6990c569d1cd318e2e1613d0a065916",
                progress_bar=progress_bar,
            )
            if downloaded is True or self._root.exists() is False:
                extract_archive(src, self._root)

        else:
            # Some sanity checks
            if self._root.exists() is False or self._root.is_dir() is False:
                raise RuntimeError("Dataset not found, try download=True to download it")

            for split_name in ["training", "validation"]:
                if self._root.joinpath(split_name).exists() is False:
                    raise RuntimeError("Dataset seems corrupted")

        super().__init__(self._root.joinpath(split), transform, target_transform, loader, is_valid_file)

    def __getitem__(self, index: int) -> tuple[str, torch.Tensor, Any]:
        (path, target) = self.samples[index]
        sample = self.loader(path)
        if self.transform is not None:
            sample = self.transform(sample)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return (path, sample, target)



"""
MobileNet v2, adapted from
https://github.com/pytorch/vision/blob/main/torchvision/models/mobilenetv2.py

Paper "MobileNetV2: Inverted Residuals and Linear Bottlenecks", https://arxiv.org/abs/1801.04381
"""

# Reference license: BSD 3-Clause

from collections import OrderedDict
from collections.abc import Callable
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation

from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import make_divisible


class InvertedResidual(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        padding: tuple[int, int],
        expansion_factor: float,
        shortcut: bool,
        activation_layer: Callable[..., nn.Module] = nn.ReLU6,
    ) -> None:
        super().__init__()
        num_expfilter = int(round(in_channels * expansion_factor))

        self.shortcut = shortcut
        self.block = nn.Sequential(
            Conv2dNormActivation(
                in_channels,
                num_expfilter,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
                activation_layer=activation_layer,
            ),
            Conv2dNormActivation(
                num_expfilter,
                num_expfilter,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                groups=num_expfilter,
                bias=False,
                activation_layer=activation_layer,
            ),
            Conv2dNormActivation(
                num_expfilter,
                out_channels,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
                activation_layer=None,
            ),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.shortcut is True:
            return x + self.block(x)

        return self.block(x)


# pylint: disable=invalid-name
class MobileNet_v2(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        alpha: float = self.config["alpha"]

        # t - expansion factor
        # c - num_filters (channels)
        # n - number of repetitions
        # s - stride
        inverted_residual_setting = [
            # t, c, n, s
            [1, 16, 1, 1],
            [6, 24, 2, 2],
            [6, 32, 3, 2],
            [6, 64, 4, 2],
            [6, 96, 3, 1],
            [6, 160, 3, 2],
            [6, 320, 1, 1],
        ]

        base = make_divisible(32 * alpha, 8)
        self.stem = Conv2dNormActivation(
            self.input_channels,
            base,
            kernel_size=(3, 3),
            stride=(2, 2),
            padding=(1, 1),
            bias=False,
            activation_layer=nn.ReLU6,
        )

        layers: list[nn.Module] = []
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        i = 0
        for t, c, n, s in inverted_residual_setting:
            if s > 1:
                stages[f"stage{i}"] = nn.Sequential(*layers)
                return_channels.append(base)
                layers = []
                i += 1

            c = make_divisible(c * alpha, 8)
            layers.append(
                InvertedResidual(
                    base,
                    c,
                    kernel_size=(3, 3),
                    stride=(s, s),
                    padding=(1, 1),
                    expansion_factor=t,
                    shortcut=False,
                )
            )
            for _ in range(1, n):
                layers.append(
                    InvertedResidual(
                        c,
                        c,
                        kernel_size=(3, 3),
                        stride=(1, 1),
                        padding=(1, 1),
                        expansion_factor=t,
                        shortcut=True,
                    )
                )

            base = c

        stages[f"stage{i}"] = nn.Sequential(*layers)
        return_channels.append(base)

        last_channels = make_divisible(1280 * max(1.0, alpha), 8)
        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            Conv2dNormActivation(
                c,
                last_channels,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
                activation_layer=nn.ReLU6,
            ),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels[1:5]
        self.embedding_size = last_channels
        self.classifier = self.create_classifier()

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out")
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config("mobilenet_v2_0_25", MobileNet_v2, config={"alpha": 0.25})
registry.register_model_config("mobilenet_v2_0_5", MobileNet_v2, config={"alpha": 0.5})
registry.register_model_config("mobilenet_v2_0_75", MobileNet_v2, config={"alpha": 0.75})
registry.register_model_config("mobilenet_v2_1_0", MobileNet_v2, config={"alpha": 1.0})
registry.register_model_config("mobilenet_v2_1_25", MobileNet_v2, config={"alpha": 1.25})
registry.register_model_config("mobilenet_v2_1_5", MobileNet_v2, config={"alpha": 1.5})
registry.register_model_config("mobilenet_v2_1_75", MobileNet_v2, config={"alpha": 1.75})
registry.register_model_config("mobilenet_v2_2_0", MobileNet_v2, config={"alpha": 2.0})

registry.register_weights(
    "mobilenet_v2_1_0_il-common",
    {
        "description": "MobileNet v2 (1.0 multiplier) model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 10.6,
                "sha256": "d6182293e98c102026f7cdc0d446aaf0e511232173c4b98c1a882c9f147be6e7",
            }
        },
        "net": {"network": "mobilenet_v2_1_0", "tag": "il-common"},
    },
)



import argparse
import logging
from typing import Any

from birder.common import cli
from birder.conf import settings
from birder.results.detection import Results

logger = logging.getLogger(__name__)


def print_report(results_dict: dict[str, Results]) -> None:
    if len(results_dict) == 1:
        results = next(iter(results_dict.values()))
        results.pretty_print()
        return

    raise NotADirectoryError


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "det-results",
        allow_abbrev=False,
        help="read and process detection result files",
        description="read and process detection result files",
        epilog=(
            "Usage examples:\n"
            "python -m birder.tools det-results "
            "results/faster_rcnn_coco_csp_resnet_50_imagenet1k_91_e0_640px_5000.json --print\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument("--print", default=False, action="store_true", help="print results table")
    subparser.add_argument("--save-summary", default=False, action="store_true", help="save results summary as csv")
    subparser.add_argument("--summary-suffix", type=str, help="add suffix to summary file")
    subparser.add_argument("result_files", type=str, nargs="+", help="result files to process")
    subparser.set_defaults(func=main)


def main(args: argparse.Namespace) -> None:
    results_dict: dict[str, Results] = {}
    for results_file in args.result_files:
        results = Results.load(results_file)
        result_name = results_file.split("/")[-1]
        results_dict[result_name] = results

    if args.print is True:
        print_report(results_dict)

    if args.save_summary is True:
        if args.summary_suffix is not None:
            summary_path = settings.RESULTS_DIR.joinpath(f"summary_{args.summary_suffix}.csv")
        else:
            summary_path = settings.RESULTS_DIR.joinpath("summary.csv")

        if summary_path.exists() is True:
            logger.warning(f"Summary already exists '{summary_path}', skipping...")
        else:
            logger.info(f"Writing results summary at '{summary_path}...")
            raise NotImplementedError



"""
EfficientNet v2, adapted from
https://github.com/pytorch/vision/blob/main/torchvision/models/efficientnet.py

Paper "EfficientNetV2: Smaller Models and Faster Training",
https://arxiv.org/abs/2104.00298
"""

# Reference license: BSD 3-Clause

import math
from collections import OrderedDict
from typing import Any
from typing import Literal
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenRetentionResultType
from birder.net.base import make_divisible
from birder.net.efficientnet_v1 import MBConv


def adjust_channels(channels: int, width: float, min_value: Optional[int] = None) -> int:
    return make_divisible(channels * width, 8, min_value)


def adjust_depth(num_layers: int, depth: float) -> int:
    return int(math.ceil(num_layers * depth))


class FusedMBConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        expand_ratio: float,
        stochastic_depth_prob: float,
    ) -> None:
        super().__init__()

        if stride == (1, 1) and in_channels == out_channels:
            self.use_res_connect = True
        else:
            self.use_res_connect = False

        layers = []
        activation_layer = nn.SiLU

        padding = ((kernel_size[0] - 1) // 2, (kernel_size[1] - 1) // 2)
        expanded_channels = adjust_channels(in_channels, expand_ratio)
        if expanded_channels != in_channels:
            # Fused Expand
            layers.append(
                Conv2dNormActivation(
                    in_channels,
                    expanded_channels,
                    kernel_size=kernel_size,
                    stride=stride,
                    padding=padding,
                    bias=False,
                    activation_layer=activation_layer,
                )
            )

            # Project
            layers.append(
                Conv2dNormActivation(
                    expanded_channels,
                    out_channels,
                    kernel_size=(1, 1),
                    stride=(1, 1),
                    padding=(0, 0),
                    bias=False,
                    activation_layer=None,
                )
            )

        else:
            layers.append(
                Conv2dNormActivation(
                    in_channels,
                    out_channels,
                    kernel_size=kernel_size,
                    stride=stride,
                    padding=padding,
                    bias=False,
                    activation_layer=activation_layer,
                )
            )

        self.block = nn.Sequential(*layers)
        self.stochastic_depth = StochasticDepth(stochastic_depth_prob, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        branch = self.block(x)
        if self.use_res_connect is True:
            return x + self.stochastic_depth(branch)

        return branch


# pylint: disable=invalid-name,too-many-locals
class EfficientNet_v2(DetectorBackbone, PreTrainEncoder, MaskedTokenRetentionMixin):
    block_group_regex = r"body\.stage(\d+)\.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        width_coefficient = 1.0
        depth_coefficient = 1.0
        last_channels: int = self.config.get("last_channels", 1280)
        expand_ratio: list[int] = self.config["expand_ratio"]
        kernel_size: list[tuple[int, int]] = self.config["kernel_size"]
        strides: list[tuple[int, int]] = self.config["strides"]
        in_channels: list[int] = self.config["in_channels"]
        out_channels: list[int] = self.config["out_channels"]
        repeats: list[int] = self.config["repeats"]
        dropout_rate: float = self.config["dropout_rate"]
        drop_path_rate: float = self.config.get("drop_path_rate", 0.2)

        self.dropout_rate = dropout_rate
        in_channels = [adjust_channels(ch, width_coefficient) for ch in in_channels]
        out_channels = [adjust_channels(ch, width_coefficient) for ch in out_channels]
        repeats = [adjust_depth(re, depth_coefficient) for re in repeats]

        self.stem = Conv2dNormActivation(
            self.input_channels,
            in_channels[0],
            kernel_size=(3, 3),
            stride=(2, 2),
            padding=(1, 1),
            bias=False,
            activation_layer=nn.SiLU,
        )

        layers: list[nn.Module] = []
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        stage_id = 0
        total_stage_blocks = sum(repeats)
        stage_block_id = 0
        for i, repeat in enumerate(repeats):
            for r in range(repeat):
                # Adjust stochastic depth probability based on the depth of the stage block
                sd_prob = drop_path_rate * float(stage_block_id) / total_stage_blocks

                if r > 0:
                    in_ch = out_channels[i]
                    stride = (1, 1)

                else:
                    in_ch = in_channels[i]
                    stride = strides[i]

                if stride[0] > 1 or stride[1] > 1:
                    stages[f"stage{stage_id}"] = nn.Sequential(*layers)
                    return_channels.append(in_ch)
                    layers = []
                    stage_id += 1

                if i < 3:
                    layers.append(
                        FusedMBConv(
                            in_ch,
                            out_channels[i],
                            kernel_size=kernel_size[i],
                            stride=stride,
                            expand_ratio=expand_ratio[i],
                            stochastic_depth_prob=sd_prob,
                        )
                    )

                else:
                    layers.append(
                        MBConv(
                            in_ch,
                            out_channels[i],
                            kernel_size=kernel_size[i],
                            stride=stride,
                            expand_ratio=expand_ratio[i],
                            stochastic_depth_prob=sd_prob,
                        )
                    )

                stage_block_id += 1

        stages[f"stage{stage_id}"] = nn.Sequential(*layers)
        return_channels.append(out_channels[-1])

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            Conv2dNormActivation(
                out_channels[-1],
                last_channels,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
                activation_layer=nn.SiLU,
            ),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels[1:5]
        self.embedding_size = last_channels
        self.classifier = self.create_classifier()

        self.stem_stride = 2
        self.stem_width = in_channels[0]
        self.encoding_size = out_channels[-1]

        # Weights initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out")
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

            elif isinstance(m, nn.Linear):
                init_range = 1.0 / math.sqrt(m.out_features)
                nn.init.uniform_(m.weight, -init_range, init_range)
                nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        x = self.stem(x)
        x = mask_tensor(x, mask, patch_factor=self.max_stride // self.stem_stride, mask_token=mask_token)
        x = self.body(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            result["features"] = x
        if return_keys in ("all", "embedding"):
            result["embedding"] = self.features(x)

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def create_classifier(self, embed_dim: Optional[int] = None) -> nn.Module:
        if self.num_classes == 0:
            return nn.Identity()

        if embed_dim is None:
            embed_dim = self.embedding_size

        return nn.Sequential(
            nn.Dropout(p=self.dropout_rate, inplace=True),
            nn.Linear(embed_dim, self.num_classes),
        )


registry.register_model_config(
    "efficientnet_v2_s",
    EfficientNet_v2,  # input_resolution = 300 / 384 (train / eval)
    config={
        "expand_ratio": [1, 4, 4, 4, 6, 6],
        "kernel_size": [(3, 3), (3, 3), (3, 3), (3, 3), (3, 3), (3, 3)],
        "strides": [(1, 1), (2, 2), (2, 2), (2, 2), (1, 1), (2, 2)],
        "in_channels": [24, 24, 48, 64, 128, 160],
        "out_channels": [24, 48, 64, 128, 160, 256],
        "repeats": [2, 4, 4, 6, 9, 15],
        "dropout_rate": 0.2,
    },
)
registry.register_model_config(
    "efficientnet_v2_m",
    EfficientNet_v2,  # input_resolution = 384 / 480 (train / eval)
    config={
        "expand_ratio": [1, 4, 4, 4, 6, 6, 6],
        "kernel_size": [(3, 3), (3, 3), (3, 3), (3, 3), (3, 3), (3, 3), (3, 3)],
        "strides": [(1, 1), (2, 2), (2, 2), (2, 2), (1, 1), (2, 2), (1, 1)],
        "in_channels": [24, 24, 48, 80, 160, 176, 304],
        "out_channels": [24, 48, 80, 160, 176, 304, 512],
        "repeats": [3, 5, 5, 7, 14, 18, 5],
        "dropout_rate": 0.3,
    },
)
registry.register_model_config(
    "efficientnet_v2_l",
    EfficientNet_v2,  # input_resolution = 384 / 480 (train / eval)
    config={
        "expand_ratio": [1, 4, 4, 4, 6, 6, 6],
        "kernel_size": [(3, 3), (3, 3), (3, 3), (3, 3), (3, 3), (3, 3), (3, 3)],
        "strides": [(1, 1), (2, 2), (2, 2), (2, 2), (1, 1), (2, 2), (1, 1)],
        "in_channels": [32, 32, 64, 96, 192, 224, 384],
        "out_channels": [32, 64, 96, 192, 224, 384, 640],
        "repeats": [4, 7, 7, 10, 19, 25, 7],
        "dropout_rate": 0.4,
    },
)
registry.register_model_config(
    "efficientnet_v2_xl",
    EfficientNet_v2,  # input_resolution = 384 / 512 (train / eval)
    config={
        "expand_ratio": [1, 4, 4, 4, 6, 6, 6],
        "kernel_size": [(3, 3), (3, 3), (3, 3), (3, 3), (3, 3), (3, 3), (3, 3)],
        "strides": [(1, 1), (2, 2), (2, 2), (2, 2), (1, 1), (2, 2), (1, 1)],
        "in_channels": [32, 32, 64, 96, 192, 256, 512],
        "out_channels": [32, 64, 96, 192, 256, 512, 640],
        "repeats": [4, 8, 8, 16, 24, 32, 8],
        "dropout_rate": 0.4,
    },
)

registry.register_weights(
    "efficientnet_v2_s_il-common",
    {
        "description": "EfficientNet v2 small model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 79.7,
                "sha256": "abd0310951ae2dcc56792473bcefab49ab28971746322cd561d6c369503e4d6f",
            }
        },
        "net": {"network": "efficientnet_v2_s", "tag": "il-common"},
    },
)
registry.register_weights(
    "efficientnet_v2_s_arabian-peninsula",
    {
        "url": "https://huggingface.co/birder-project/efficientnet_v2_s_arabian-peninsula/resolve/main",
        "description": "EfficientNet v2 small model trained on the arabian-peninsula dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 81.5,
                "sha256": "ad708a255d2e0fbcacc907d0b42a3881180a4e0a9b238dec409985b502dac62d",
            }
        },
        "net": {"network": "efficientnet_v2_s", "tag": "arabian-peninsula"},
    },
)



"""
Maximum Manifold Capacity Representations (MMCR), adapted from
https://github.com/ThomasYerxa/mmcr/blob/master/mmcr/imagenet/models.py

Paper "Learning Efficient Coding of Natural Images with Maximum Manifold Capacity Representations",
https://arxiv.org/abs/2303.03307
"""

# Reference license: MIT

import copy
from typing import Any
from typing import Optional

import torch
import torch.distributed as dist
import torch.nn.functional as F
from torch import nn

from birder.common import training_utils
from birder.net.base import BaseNet
from birder.net.ssl.base import SSLBaseNet


class MMCRMomentumLoss(nn.Module):
    def __init__(self, lambda_coeff: float, n_aug: int) -> None:
        super().__init__()
        self.lambda_coeff = lambda_coeff
        self.n_aug = n_aug

    def forward(self, z: torch.Tensor, z_m: torch.Tensor) -> torch.Tensor:
        z = F.normalize(z, dim=-1)
        z_m = F.normalize(z_m, dim=-1)

        z_local_ = z.reshape(z.size(0) // self.n_aug, self.n_aug, -1).permute(0, 2, 1).contiguous()
        z_local_m = z_m.reshape(z_m.size(0) // self.n_aug, self.n_aug, -1).permute(0, 2, 1).contiguous()

        # Gather across devices into list
        if training_utils.is_dist_available_and_initialized() is True:
            z_list = [torch.zeros_like(z_local_) for _ in range(dist.get_world_size())]
            dist.all_gather(z_list, z_local_, async_op=False)
            z_list[dist.get_rank()] = z_local_

            # Gather momentum outputs
            z_m_list = [torch.zeros_like(z_local_m) for _ in range(dist.get_world_size())]
            dist.all_gather(z_m_list, z_local_m, async_op=False)
            z_m_list[dist.get_rank()] = z_local_m

            # Append all
            z_local = torch.concat(z_list, dim=0)
            z_m_local = torch.concat(z_m_list, dim=0)

        else:
            z_local = z_local_
            z_m_local = z_local_m

        if self.lambda_coeff == 0:
            local_nuc = 0
        else:
            local_nuc = torch.linalg.svdvals(z_local).sum()  # pylint: disable=not-callable

        centroids = (torch.mean(z_local, dim=-1) + torch.mean(z_m_local, dim=-1)) * 0.5

        # Filter infs and NaNs
        selected = centroids.isfinite().all(dim=1)
        centroids = centroids[selected]

        if selected.sum() != centroids.size(0):
            # NaN filtered
            pass

        global_nuc = torch.linalg.svdvals(centroids).sum()  # pylint: disable=not-callable

        loss = -1.0 * global_nuc + self.lambda_coeff * local_nuc / z_local.size(0)

        return loss


class MMCREncoder(nn.Module):
    def __init__(self, backbone: BaseNet, projector_dims: list[int]) -> None:
        super().__init__()
        self.backbone = backbone
        sizes = [self.backbone.embedding_size] + projector_dims

        # Projection head
        layers = []
        for i in range(len(sizes) - 2):
            layers.append(nn.Linear(sizes[i], sizes[i + 1], bias=False))
            layers.append(nn.BatchNorm1d(sizes[i + 1]))
            layers.append(nn.ReLU(inplace=True))

        layers.append(nn.Linear(sizes[-2], sizes[-1], bias=False))
        self.projector = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.backbone.embedding(x)
        x = torch.flatten(x, start_dim=1)
        x = self.projector(x)

        return x


class MMCR(SSLBaseNet):
    def __init__(
        self,
        backbone: BaseNet,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(backbone, config=config, size=size)
        assert self.config is not None, "must set config"

        projector_dims: list[int] = self.config["projector_dims"]

        momentum_backbone = copy.deepcopy(self.backbone)

        encoder = MMCREncoder(self.backbone, projector_dims=projector_dims)
        momentum_encoder = MMCREncoder(momentum_backbone, projector_dims=projector_dims)

        self.encoder = encoder
        self.momentum_encoder = momentum_encoder

        # Weights initialization
        self.momentum_encoder.load_state_dict(self.encoder.state_dict())

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        (C, H, W) = x.shape[-3:]  # B, num_views, C, H, W
        x = x.reshape(-1, C, H, W)
        z = self.encoder(x)
        z_m = self.momentum_encoder(x)

        return (z, z_m)



"""
EdgeViT, adapted from
https://github.com/saic-fi/edgevit/blob/master/src/edgevit.py

Paper "EdgeViTs: Competing Light-weight CNNs on Mobile Devices with Vision Transformers",
https://arxiv.org/abs/2205.03436
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from collections.abc import Callable
from functools import partial
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import StochasticDepth

from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class MLP(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_features: int,
        out_features: int,
        act_layer: Callable[..., nn.Module] = nn.GELU,
    ) -> None:
        super().__init__()
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.fc1(x)
        x = self.act(x)
        x = self.fc2(x)

        return x


class ConvMLP(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_features: int,
        out_features: int,
        act_layer: Callable[..., nn.Module] = nn.GELU,
    ) -> None:
        super().__init__()
        self.fc1 = nn.Conv2d(in_features, hidden_features, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=True)
        self.act = act_layer()
        self.fc2 = nn.Conv2d(
            hidden_features, out_features, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=True
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.fc1(x)
        x = self.act(x)
        x = self.fc2(x)

        return x


class GlobalSparseAttn(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        sr_ratio: int,
    ) -> None:
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3)
        self.proj = nn.Linear(dim, dim)

        self.sr = sr_ratio
        if self.sr > 1:
            self.sampler = nn.AvgPool2d(kernel_size=(1, 1), stride=(self.sr, self.sr), padding=(0, 0))
            self.local_prop = nn.ConvTranspose2d(
                dim,
                dim,
                kernel_size=(self.sr, self.sr),
                stride=(self.sr, self.sr),
                padding=(0, 0),
                output_padding=(0, 0),
                groups=dim,
                bias=True,
            )
            self.norm = nn.LayerNorm(dim)

        else:
            self.local_prop = nn.Identity()
            self.sampler = nn.Identity()
            self.norm = nn.Identity()

    def forward(self, x: torch.Tensor, H: int, W: int) -> torch.Tensor:
        (B, _, C) = x.size()  # B, N, C
        if self.sr > 1:
            x = x.transpose(1, 2).reshape(B, C, H, W)
            x = self.sampler(x)
            x = x.flatten(2).transpose(1, 2)

        qkv = self.qkv(x).reshape(B, -1, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q = qkv[0]
        k = qkv[1]
        v = qkv[2]

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)

        x = (attn @ v).transpose(1, 2).reshape(B, -1, C)

        if self.sr > 1:
            x = x.permute(0, 2, 1).reshape(B, C, int(H / self.sr), int(W / self.sr))
            x = self.local_prop(x)
            x = x.reshape(B, C, -1).permute(0, 2, 1)
            x = self.norm(x)

        x = self.proj(x)

        return x


class LocalAgg(nn.Module):
    def __init__(
        self,
        dim: int,
        mlp_ratio: float,
        act_layer: Callable[..., nn.Module],
        drop_path: float,
    ) -> None:
        super().__init__()
        self.pos_embed = nn.Conv2d(dim, dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=dim, bias=True)
        self.norm1 = nn.BatchNorm2d(dim)
        self.conv1 = nn.Conv2d(dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=True)
        self.conv2 = nn.Conv2d(dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=True)
        self.attn = nn.Conv2d(dim, dim, kernel_size=(5, 5), stride=(1, 1), padding=(2, 2), groups=dim, bias=True)
        self.drop_path = StochasticDepth(drop_path, mode="row")
        self.norm2 = nn.BatchNorm2d(dim)
        self.mlp = ConvMLP(in_features=dim, hidden_features=int(dim * mlp_ratio), out_features=dim, act_layer=act_layer)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.pos_embed(x)
        x = x + self.drop_path(self.conv2(self.attn(self.conv1(self.norm1(x)))))
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        return x


class SelfAttn(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        mlp_ratio: float,
        act_layer: Callable[..., nn.Module],
        norm_layer: Callable[..., nn.Module],
        drop_path: float,
        sr_ratio: int,
    ) -> None:
        super().__init__()
        self.pos_embed = nn.Conv2d(dim, dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=dim, bias=True)
        self.norm1 = norm_layer(dim)
        self.attn = GlobalSparseAttn(dim, num_heads=num_heads, sr_ratio=sr_ratio)
        self.drop_path = StochasticDepth(drop_path, mode="row")
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = MLP(in_features=dim, hidden_features=mlp_hidden_dim, out_features=dim, act_layer=act_layer)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.pos_embed(x)
        (B, N, H, W) = x.size()
        x = x.flatten(2).transpose(1, 2)
        x = x + self.drop_path(self.attn(self.norm1(x), H, W))
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        x = x.transpose(1, 2).reshape(B, N, H, W)

        return x


class LGLBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        mlp_ratio: float,
        act_layer: Callable[..., nn.Module],
        norm_layer: Callable[..., nn.Module],
        drop_path: float,
        sr_ratio: int,
    ):
        super().__init__()

        if sr_ratio > 1:
            self.local_agg = LocalAgg(dim, mlp_ratio, act_layer, drop_path)

        else:
            self.local_agg = nn.Identity()

        self.self_attn = SelfAttn(dim, num_heads, mlp_ratio, act_layer, norm_layer, drop_path, sr_ratio)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.local_agg(x)
        x = self.self_attn(x)
        return x


class PatchEmbed(nn.Module):
    def __init__(self, patch_size: int, in_channels: int, embed_dim: int) -> None:
        super().__init__()
        self.proj = nn.Conv2d(
            in_channels,
            embed_dim,
            kernel_size=(patch_size, patch_size),
            stride=(patch_size, patch_size),
            padding=(0, 0),
            bias=True,
        )
        self.norm = nn.LayerNorm(embed_dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.proj(x)
        (B, _, H, W) = x.size()
        x = x.flatten(2).transpose(1, 2)
        x = self.norm(x)
        x = x.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()

        return x


class EdgeViTSage(nn.Module):
    def __init__(
        self,
        patch_size: int,
        in_channels: int,
        embed_dim: int,
        depth: int,
        num_heads: int,
        mlp_ratio: float,
        act_layer: Callable[..., nn.Module],
        norm_layer: Callable[..., nn.Module],
        drop_path: list[float],
        sr_ratio: int,
    ) -> None:
        super().__init__()
        self.patch_embed = PatchEmbed(patch_size=patch_size, in_channels=in_channels, embed_dim=embed_dim)
        layers = []
        for i in range(depth):
            layers.append(
                LGLBlock(
                    dim=embed_dim,
                    num_heads=num_heads,
                    mlp_ratio=mlp_ratio,
                    act_layer=act_layer,
                    norm_layer=norm_layer,
                    drop_path=drop_path[i],
                    sr_ratio=sr_ratio,
                )
            )
        self.blocks = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.patch_embed(x)
        x = self.blocks(x)

        return x


class EdgeViT(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        patch_size = [4, 2, 2, 2]
        mlp_ratio = [4.0, 4.0, 4.0, 4.0]
        sr_ratios = [4, 2, 2, 1]
        depth: list[int] = self.config["depth"]
        embed_dim: list[int] = self.config["embed_dim"]
        head_dim: int = self.config["head_dim"]
        drop_path_rate: float = self.config.get("drop_path_rate", 0.1)

        num_stages = len(depth)
        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depth)).split(depth)]
        num_heads = [dim // head_dim for dim in embed_dim]
        norm_layer = partial(nn.LayerNorm, eps=1e-6)

        prev_dim = self.input_channels
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(num_stages):
            stages[f"stage{i+1}"] = EdgeViTSage(
                patch_size=patch_size[i],
                in_channels=prev_dim,
                embed_dim=embed_dim[i],
                depth=depth[i],
                num_heads=num_heads[i],
                mlp_ratio=mlp_ratio[i],
                act_layer=nn.GELU,
                norm_layer=norm_layer,
                drop_path=dpr[i],
                sr_ratio=sr_ratios[i],
            )
            return_channels.append(embed_dim[i])
            prev_dim = embed_dim[i]

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.BatchNorm2d(embed_dim[-1]),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = embed_dim[-1]
        self.classifier = self.create_classifier()

        # Weights initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

            elif isinstance(m, nn.LayerNorm):
                nn.init.constant_(m.bias, 0)
                nn.init.constant_(m.weight, 1.0)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def create_classifier(self, embed_dim: Optional[int] = None) -> nn.Module:
        if self.num_classes == 0:
            return nn.Identity()

        if embed_dim is None:
            embed_dim = self.embedding_size

        return nn.Linear(embed_dim, self.num_classes, bias=False)


registry.register_model_config(
    "edgevit_xxs", EdgeViT, config={"depth": [1, 1, 3, 2], "embed_dim": [36, 72, 144, 288], "head_dim": 36}
)
registry.register_model_config(
    "edgevit_xs", EdgeViT, config={"depth": [1, 1, 3, 1], "embed_dim": [48, 96, 240, 384], "head_dim": 48}
)
registry.register_model_config(
    "edgevit_s", EdgeViT, config={"depth": [1, 2, 5, 3], "embed_dim": [48, 96, 240, 384], "head_dim": 48}
)

registry.register_weights(
    "edgevit_xxs_il-common",
    {
        "description": "EdgeViT XXS model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 14.9,
                "sha256": "b4ae9c4f871bafc47c757b65ee8f7ce2dc427fff55171f4feaafdbb5ba02ff3f",
            }
        },
        "net": {"network": "edgevit_xxs", "tag": "il-common"},
    },
)
registry.register_weights(
    "edgevit_xs_il-common",
    {
        "description": "EdgeViT XS model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 25.0,
                "sha256": "38ab9c6f10586a84c2abaa55919cba56cbd2196ddf771befa09eb71c79d3634c",
            }
        },
        "net": {"network": "edgevit_xs", "tag": "il-common"},
    },
)
